-- ===========================================================================
--	Contains scaffolding for WorldRankings and other Right Anchored screens
-- ===========================================================================
include("TabSupport");
include("InstanceManager");
include("SupportFunctions");
include("AnimSidePanelSupport");

-- ===========================================================================
--	DEBUG
-- ===========================================================================
local m_isDebugForceShowAllScoreCategories :boolean = false;		-- (false) Show all scoring categories under details

-- ===========================================================================
--	CONSTANTS
-- ===========================================================================
local RELOAD_CACHE_ID:string = "WorldRankings"; -- Must be unique (usually the same as the file name)
local REQUIREMENT_CONTEXT:string = "VictoryProgress";
local DATA_FIELD_SELECTION:string = "Selection";
local DATA_FIELD_LEADER_TOOLTIP:string = "LeaderTooltip";
local DATA_FIELD_HEADER_HEIGHT:string = "HeaderHeight";
local DATA_FIELD_HEADER_RESIZED:string = "HeaderResized";
local DATA_FIELD_HEADER_EXPANDED:string = "HeaderExpanded";
local DATA_FIELD_OVERALL_PLAYERS_IM:string = "OverallPlayersIM";
local DATA_FIELD_DOMINATED_CITIES_IM:string = "DominatedCitiesIM";
local DATA_FIELD_RELIGION_CONVERTED_CIVS_IM:string = "ConvertedCivsIM";

local PADDING_HEADER:number = 10;
local PADDING_CULTURE_HEADER:number = 90;
local PADDING_GENERIC_ITEM_BG:number = 25;
local PADDING_TAB_BUTTON_TEXT:number = 17;
local PADDING_EXTRA_TAB_BG:number = 10;
local PADDING_EXTRA_TAB_SHADOW:number = 23;
local PADDING_ADVISOR_TEXT_BG:number = 20;
local PADDING_RELIGION_NAME_BG:number = 42;
local PADDING_RELIGION_BG_HEIGHT:number = 26;
local PADDING_VICTORY_GRADIENT:number = 45;
local PADDING_NEXT_STEP_HIGHLIGHT:number = 4;
local PADDING_VICTORY_LABEL_UNDERLINE:number = 90;
local PADDING_SCORE_DETAILS_BUTTON_WIDTH:number = 40;
local OFFSET_VIEW_CONTENTS:number = 130;
local OFFSET_ADVISOR_ICON_Y:number = 5;
local OFFSET_ADVISOR_TEXT_Y:number = 70;
local OFFSET_HIDDEN_SCROLLBAR:number = 7;
local OFFSET_CONTRACT_BUTTON_Y:number = 63;
local OFFSET_SCIENCE_REQUIREMENTS_Y:number = 80;
local SIZE_OVERALL_TOP_PLAYER_ICON:number = 48;
local SIZE_OVERALL_PLAYER_ICON:number = 36;
local SIZE_OVERALL_BG_HEIGHT:number = 100;
local SIZE_OVERALL_INSTANCE:number = 40;
local SIZE_VICTORY_ICON_SMALL:number = 64;
local SIZE_RELIGION_BG_HEIGHT:number = 55;
local SIZE_RELIGION_ICON_SMALL:number = 22;
local SIZE_RELIGION_CIV_ICON:number = 30;
local SIZE_GENERIC_ITEM_MIN_Y:number = 54;
local SIZE_SCORE_ITEM_DEFAULT:number = 54;
local SIZE_SCORE_ITEM_DETAILS:number = 180;
local SIZE_STACK_DEFAULT:number = 225;
local SIZE_HEADER_DEFAULT:number = 60;
local SIZE_HEADER_MIN_Y:number = 46;
local SIZE_HEADER_MAX_Y:number = 270;
local SIZE_HEADER_ICON:number = 80;
local SIZE_LEADER_ICON:number = 55;
local SIZE_CIV_ICON:number = 36;

local TAB_SCORE:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCORE_TAB");
local TAB_OVERALL:string = Locale.Lookup("LOC_WORLD_RANKINGS_OVERALL_TAB");
local TAB_SCIENCE:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_TAB");
local TAB_CULTURE:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_TAB");
local TAB_RELIGION:string = Locale.Lookup("LOC_WORLD_RANKINGS_RELIGION_TAB");
local TAB_DOMINATION:string = Locale.Lookup("LOC_WORLD_RANKINGS_DOMINATION_TAB");
local TAB_ECONOMIC:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_TAB");
local TAB_METROPOLIS:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_TAB");
local TAB_TERRITORIAL:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_TAB");
local TAB_DIPLOMATIC:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_TAB");
--local TAB_INDUSTRIAL:string = Locale.Lookup("LOC_WORLD_RANKINGS_INDUSTRIAL_TAB");
--local TAB_ENVIRONMENTAL:string = Locale.Lookup("LOC_WORLD_RANKINGS_ENVIRONMENTAL_TAB");
--local TAB_COLONIAL:string = Locale.Lookup("LOC__WORLD_RANKINGS_COLONIAL_TAB");

local SCORE_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCORE_VICTORY");
local SCORE_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCORE_DETAILS");

local SCIENCE_ICON:string = "ICON_VICTORY_TECHNOLOGY";
local SCIENCE_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_VICTORY");
local SCIENCE_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_DETAILS");
local SCIENCE_REQUIREMENTS:table = {
	Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_REQUIREMENT_1"),
	Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_REQUIREMENT_2"),
	Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_REQUIREMENT_3")
};

local CULTURE_ICON:string = "ICON_VICTORY_CULTURE";
local CULTURE_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_VICTORY");
local CULTURE_VICTORY_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_VICTORY_DETAILS");
local CULTURE_DOMESTIC_TOURISTS:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_DETAILS_DOMESTIC_TOURISTS");
local CULTURE_VISITING_TOURISTS:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_DETAILS_VISITING_TOURISTS");

local DOMINATION_ICON:string = "ICON_VICTORY_DOMINATION";
local DOMINATION_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_DOMINATION_VICTORY");
local DOMINATION_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_DOMINATION_DETAILS");
local DOMINATION_HAS_ORIGINAL_CAPITAL:string = Locale.Lookup("LOC_WORLD_RANKINGS_DOMINATION_HAS_ORIGINAL_CAPITAL");

local RELIGION_ICON:string = "ICON_VICTORY_RELIGIOUS";
local RELIGION_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_RELIGION_VICTORY");
local RELIGION_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_RELIGION_DETAILS");

local ICON_GENERIC:string = "ICON_VICTORY_GENERIC";
local ICON_UNKNOWN_CIV:string = "ICON_CIVILIZATION_UNKNOWN";
local LOC_UNKNOWN_CIV:string = Locale.Lookup("LOC_WORLD_RANKING_UNMET_PLAYER");
local LOC_UNKNOWN_CIV_COLORED:string = Locale.Lookup("LOC_WORLD_RANKING_UNMET_PLAYER_COLORED");

local UNKNOWN_COLOR:number = RGBAValuesToABGRHex(1, 1, 1, 1);

--antonjs: Removed the other state and related text, in favor of showing all information together. Leaving state functionality intact in case we want to use it in the future.
--[[
local CULTURE_HOW_TO_VICTORY:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_HOW_TO_VICTORY");
local CULTURE_HOW_TO_TOURISM:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_HOW_TO_TOURISM");
local CULTURE_TOURISM_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_CULTURE_TOURISM_DETAILS");
--]]
local CULTURE_HEADER_STATES:table = {
	WHAT_IS_CULTURE_VICTORY	= 0;
};

local SPACE_PORT_DISTRICT_INFO:table = GameInfo.Districts["DISTRICT_SPACEPORT"];
local EARTH_SATELLITE_PROJECT_INFOS:table = {
	GameInfo.Projects["PROJECT_LAUNCH_EARTH_SATELLITE"]
};
local MOON_LANDING_PROJECT_INFOS:table = {
	GameInfo.Projects["PROJECT_LAUNCH_MOON_LANDING"]
};
local MARS_COLONY_PROJECT_INFOS:table = { 
	GameInfo.Projects["PROJECT_LAUNCH_MARS_REACTOR"],
	GameInfo.Projects["PROJECT_LAUNCH_MARS_HABITATION"],
	GameInfo.Projects["PROJECT_LAUNCH_MARS_HYDROPONICS"]
};
local SCIENCE_PROJECTS:table = {
	EARTH_SATELLITE_PROJECT_INFOS,
	MOON_LANDING_PROJECT_INFOS,
	MARS_COLONY_PROJECT_INFOS
};

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Economic
-- // ------------------------------------------------------------------------------------
local ECONOMIC_ICON:string = "ICON_VICTORY_ECONOMIC";
local ECONOMIC_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_VICTORY");
local ECONOMIC_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_DETAILS");
local ECONOMIC_DETAILS_2:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_DETAILS_2");
local ECONOMIC_REQUIREMENT_NATIONAL_BANK:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_NATIONAL_BANK");
local ECONOMIC_REQUIREMENT_GPT:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_GPT");
local ECONOMIC_REQUIREMENT_TRADE_ROUTES:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_TRADE_ROUTES");
local ECONOMIC_REQUIREMENT_LUXURY_RESOURCES:string = Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_LUXURY_RESOURCES");

local ECONOMIC_DISTRICT_INFO:table = GameInfo.Districts["DISTRICT_COMMERCIAL_HUB"];
local NATIONAL_BANK_PROJECT_INFO:table = {
	GameInfo.Projects["PROJECT_NATIONAL_BANK"]
};

local ECONOMIC_PROJECTS:table = {
	NATIONAL_BANK_PROJECT_INFO
};

local ECONOMIC_PANEL_DEFAULT_HEIGHT:number = 54;
local ECONOMIC_PANEL_SECTION_TEXT_HEIGHT:number = 18;
local ECONOMIC_PANEL_ICON_ROW_HEIGHT:number = 40;
local ECONOMIC_PANEL_ICON_ROW_BUFFER:number = 6;

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Metropolis
-- // ------------------------------------------------------------------------------------
local METROPOLIS_ICON:string = "ICON_VICTORY_METROPOLIS";
local METROPOLIS_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_VICTORY");
local METROPOLIS_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_DETAILS");
local METROPOLIS_DETAILS_2:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_DETAILS_2");
local METROPOLIS_REQUIREMENT_HIGH_POP_CITY:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_REQUIREMENT_HIGH_POP_CITY");
local METROPOLIS_REQUIREMENT_HAPPY:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_REQUIREMENT_HAPPY");
local METROPOLIS_REQUIREMENT_MONUMENT:string = Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_REQUIREMENT_MONUMENT");

local METROPOLIS_POPULATION_TARGET:number = 50;
local METROPOLIS_MIN_HAPPINESS:number = 5; -- HAPPINESS_HAPPY (6 in database - 1 as array is 0 based in lua)
local METROPOLIS_MIN_AI_HAPPINESS:number = 4; -- HAPPINESS_CONTENT (5 in database - 1 as array is 0 based in lua)

-- TODO: move to Global_Parameters in database
--local METROPOLIS_POPULATION_TARGET:number = tonumber(GameInfo.GlobalParameters["METROPOLIS_POPULATION_TARGET"].Value);
--local METROPOLIS_MIN_HAPPINESS:number = tonumber(GameInfo.GlobalParameters["METROPOLIS_MIN_HAPPINESS"].Value);

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Territorial
-- // ------------------------------------------------------------------------------------
local TERRITORIAL_ICON:string = "ICON_VICTORY_TERRITORIAL";
local TERRITORIAL_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_VICTORY");
local TERRITORIAL_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_DETAILS");
local TERRITORIAL_DETAILS_2:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_DETAILS_2");
local TERRITORIAL_REQUIREMENT_MOST_TERRITORY:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_REQUIREMENT_MOST_TERRITORY");
local TERRITORIAL_REQUIREMENT_LARGEST_MILITARY:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_REQUIREMENT_LARGEST_MILITARY");
local TERRITORIAL_REQUIREMENT_TELECOMMUNICATIONS:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_REQUIREMENT_HAS_TELECOMMUNICATIONS");
local TERRITORIAL_REQUIREMENT_MONUMENT:string = Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_REQUIREMENT_MONUMENT");

local TERRITORIAL_MIN_EMPIRE_SIZE = 35; -- minimum percentage of the total claimed territory that is needed for a Territorial victory

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Diplomatic
-- // ------------------------------------------------------------------------------------
local DIPLOMATIC_ICON:string = "ICON_VICTORY_DIPLOMATIC";
local DIPLOMATIC_TITLE:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_VICTORY");
local DIPLOMATIC_DETAILS:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_DETAILS");
local DIPLOMATIC_REQUIREMENT_MAJOR_PEACE:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_MAJOR_PEACE");
local DIPLOMATIC_REQUIREMENT_MAJOR_EMBASSY:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_MAJOR_EMBASSY");
local DIPLOMATIC_REQUIREMENT_MAJOR_ALLY:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_MAJOR_ALLY");
local DIPLOMATIC_REQUIREMENT_MINOR_PEACE:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_MINOR_PEACE");
local DIPLOMATIC_REQUIREMENT_MINOR_SUZERAIN:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_MINOR_SUZERAIN");
local DIPLOMATIC_REQUIREMENT_WORLD_COUNCIL:string = Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_WORLD_COUNCIL");

local DIPLOMATIC_PANEL_DEFAULT_HEIGHT:number = 54;
local DIPLOMATIC_PANEL_SECTION_TEXT_HEIGHT:number = 18;
local DIPLOMATIC_PANEL_ICON_ROW_HEIGHT:number = 40;
local DIPLOMATIC_PANEL_ICON_ROW_BUFFER:number = 6;

local DIPLOMATIC_MINIMUM_ALLIES_REQUIRED:number = 2;
local DIPLOMATIC_MINIMUM_SUZERAIN_REQUIRED:number = 3;

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Industrial
-- // ------------------------------------------------------------------------------------

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Environmental
-- // ------------------------------------------------------------------------------------

-- // ------------------------------------------------------------------------------------
-- // CiVI Reformation Victories - Maritime\Colonial
-- // ------------------------------------------------------------------------------------

-- // ------------------------------------------------------------------------------------

local STANDARD_VICTORY_TYPES:table = {
	"VICTORY_DEFAULT",
	"VICTORY_SCORE",
	"VICTORY_TECHNOLOGY",
	"VICTORY_CULTURE",
	"VICTORY_CONQUEST",
	"VICTORY_RELIGIOUS",
	"VICTORY_ECONOMIC",
	"VICTORY_METROPOLIS",
	"VICTORY_TERRITORIAL",
	"VICTORY_DIPLOMATIC"
	--"VICTORY_INDUSTRIAL"
	--"VICTORY_ENVIRONMENTAL"
	--"VICTORY_MARITIME"
};

function IsCustomVictoryType(victoryType:string)
	for _, checkVictoryType in ipairs(STANDARD_VICTORY_TYPES) do
		if victoryType == checkVictoryType then
			return false;
		end
	end
	return true;
end

-- ===========================================================================
--	PLAYER VARIABLES
-- ===========================================================================
local m_LocalPlayer:table;
local m_LocalPlayerID:number;
-- ===========================================================================
--	SCREEN VARIABLES
-- ===========================================================================
local m_TabSupport:table; -- TabSupport
local m_AnimSupport:table; --AnimSidePanelSupport
local m_ActiveHeader:table;
local m_TotalTabSize:number = 0;
local m_MaxExtraTabSize:number = 0;
local m_ExtraTabs:table = {};
local m_HeaderInstances:table = {};
local m_ActiveViewUpdate:ifunction;
local m_ShowScoreDetails:boolean = false;
local m_CultureHeaderState:number = CULTURE_HEADER_STATES.WHAT_IS_CULTURE_VICTORY;
local m_TabSupportIM:table = InstanceManager:new("TabInstance", "Button", Controls.TabContainer);
local m_GenericHeaderIM:table = InstanceManager:new("GenericHeaderInstance", "HeaderTop"); -- Used by Score, Religion and Domination Views
local m_ScienceHeaderIM:table = InstanceManager:new("ScienceHeaderInstance", "HeaderTop", Controls.ScienceViewHeader);
local m_CultureHeaderIM:table = InstanceManager:new("CultureHeaderInstance", "HeaderTop", Controls.CultureViewHeader);
local m_OverallIM:table = InstanceManager:new("OverallInstance", "ButtonBG", Controls.OverallViewStack);
local m_ScoreIM:table = InstanceManager:new("ScoreInstance", "ButtonBG", Controls.ScoreViewStack);
local m_ScienceIM:table = InstanceManager:new("ScienceInstance", "ButtonBG", Controls.ScienceViewStack);
local m_CultureIM:table = InstanceManager:new("CultureInstance", "ButtonBG", Controls.CultureViewStack);
local m_DominationIM:table = InstanceManager:new("DominationInstance", "ButtonBG", Controls.DominationViewStack);
local m_ReligionIM:table = InstanceManager:new("ReligionInstance", "ButtonBG", Controls.ReligionViewStack);
local m_GenericIM:table = InstanceManager:new("GenericInstance", "ButtonBG", Controls.GenericViewStack);
local m_ExtraTabsIM:table = InstanceManager:new("ExtraTab", "Button", Controls.ExtraTabStack);

-- // ---------------------------------------------------------------------------------------
-- // CiVI Reformation Victories variables
-- // ---------------------------------------------------------------------------------------
local m_EconomicHeaderIM:table = InstanceManager:new("EconomicHeaderInstance", "HeaderTop", Controls.EconomicViewHeader);
local m_EconomicIM:table = InstanceManager:new("EconomicInstance", "ButtonBG", Controls.EconomicViewStack);
local m_MetropolisHeaderIM:table = InstanceManager:new("MetropolisHeaderInstance", "HeaderTop", Controls.MetropolisViewHeader);
local m_MetropolisIM:table = InstanceManager:new("MetropolisInstance", "ButtonBG", Controls.MetropolisViewStack);
local m_TerritorialHeaderIM:table = InstanceManager:new("TerritorialHeaderInstance", "HeaderTop", Controls.TerritorialViewHeader);
local m_TerritorialIM:table = InstanceManager:new("TerritorialInstance", "ButtonBG", Controls.TerritorialViewStack);
local m_DiplomaticHeaderIM:table = InstanceManager:new("DiplomaticHeaderInstance", "HeaderTop", Controls.DiplomaticViewHeader);
local m_DiplomaticIM:table = InstanceManager:new("DiplomaticInstance", "ButtonBG", Controls.DiplomaticViewStack);

local m_MapLuxuryResources:table = {};
local m_MapLuxuryResourcesCount:number = 0;
local m_LocalPlayerRequiredLuxuryResources = {};

-- // ---------------------------------------------------------------------------------------

local m_CivTooltip = {};
TTManager:GetTypeControlTable("CivTooltip", m_CivTooltip);

-- ===========================================================================
--	Called once during Init
-- ===========================================================================
function PopulateTabs()

	-- Clean up previous data
	m_ExtraTabs = {};
	m_TotalTabSize = 0;
	m_MaxExtraTabSize = 0;
	m_ExtraTabsIM:ResetInstances();
	m_TabSupportIM:ResetInstances();
	
	-- Deselect previously selected tab
	if(m_TabSupport ~= nil) then
		m_TabSupport.SelectTab(nil);
		if(m_TabSupport.prevSelectedControl ~= nil) then
			m_TabSupport.prevSelectedControl[DATA_FIELD_SELECTION]:SetHide(true);
		end
	end

	-- Create TabSupport object
	m_TabSupport = CreateTabs(Controls.TabContainer, 42, 34, 0xFF331D05);

	local defaultTab = AddTab(TAB_OVERALL, ViewOverall);

	-- Add default victory types in a pre-determined order
	if(GameConfiguration.IsAnyMultiplayer() or Game.IsVictoryEnabled("VICTORY_SCORE")) then
		AddTab(TAB_SCORE, ViewScore);
	end
	if(Game.IsVictoryEnabled("VICTORY_TECHNOLOGY")) then
		AddTab(TAB_SCIENCE, ViewScience);
	end
	if(Game.IsVictoryEnabled("VICTORY_CULTURE")) then
		AddTab(TAB_CULTURE, ViewCulture);
	end
	if(Game.IsVictoryEnabled("VICTORY_CONQUEST")) then
		AddTab(TAB_DOMINATION, ViewDomination);
	end
	if(Game.IsVictoryEnabled("VICTORY_RELIGIOUS")) then
		AddTab(TAB_RELIGION, ViewReligion);
	end
	if(Game.IsVictoryEnabled("VICTORY_ECONOMIC")) then
		AddTab(TAB_ECONOMIC, ViewEconomic);
	end
	if(Game.IsVictoryEnabled("VICTORY_METROPOLIS")) then
		AddTab(TAB_METROPOLIS, ViewMetropolis);
	end
	if(Game.IsVictoryEnabled("VICTORY_TERRITORIAL")) then
		AddTab(TAB_TERRITORIAL, ViewTerritorial);
	end
	if(Game.IsVictoryEnabled("VICTORY_DIPLOMATIC")) then
		AddTab(TAB_DIPLOMATIC, ViewDiplomatic);
	end

	-- Add custom (modded) victory types
	for row in GameInfo.Victories() do
		local victoryType:string = row.VictoryType;
		if IsCustomVictoryType(victoryType) and Game.IsVictoryEnabled(victoryType) then
			AddTab(Locale.Lookup(row.Name), function() ViewGeneric(victoryType); end);
		end
	end

	if m_TotalTabSize > Controls.TabContainer:GetSizeX() then
		Controls.ExpandExtraTabs:SetHide(false);
		for _, tabInst in pairs(m_ExtraTabs) do
			tabInst.Button:SetSizeX(m_MaxExtraTabSize);
		end
	else
		Controls.ExpandExtraTabs:SetHide(true);
	end

	Controls.ExtraTabs:SetOffsetX(-1 * Controls.ExtraTabStack:GetSizeX() + PADDING_EXTRA_TAB_BG);
	Controls.ExtraTabsBG:SetSizeVal(Controls.ExtraTabStack:GetSizeX() + PADDING_EXTRA_TAB_BG, Controls.ExtraTabStack:GetSizeY() + PADDING_EXTRA_TAB_BG);
	Controls.ExtraTabsShadow:SetSizeVal(Controls.ExtraTabStack:GetSizeX() + PADDING_EXTRA_TAB_SHADOW, Controls.ExtraTabStack:GetSizeY() + PADDING_EXTRA_TAB_SHADOW);
	
	m_TabSupport.SelectTab(defaultTab);
	m_TabSupport.EvenlySpreadTabs();
end

function AddTab(label:string, onClickCallback:ifunction)

	local tabInst:table = m_TabSupportIM:GetInstance();
	tabInst.Button[DATA_FIELD_SELECTION] = tabInst.Selection;

	tabInst.Button:SetText(label);
	local textControl = tabInst.Button:GetTextControl();
	textControl:SetHide(false);

	local textSize:number = textControl:GetSizeX();
	tabInst.Button:SetSizeX(textSize + PADDING_TAB_BUTTON_TEXT);
	tabInst.Button:RegisterCallback(Mouse.eMouseEnter,	function() UI.PlaySound("Main_Menu_Mouse_Over"); end);
	tabInst.Selection:SetSizeX(textSize + PADDING_TAB_BUTTON_TEXT + 4);

	m_TotalTabSize = m_TotalTabSize + tabInst.Button:GetSizeX();
	if m_TotalTabSize > Controls.TabContainer:GetSizeX() then -- change here to make all victories show as extra tabs
		m_TabSupportIM:ReleaseInstance(tabInst);
		AddExtraTab(label, onClickCallback);
	else

		local callback = function()
			if(m_TabSupport.prevSelectedControl ~= nil) then
				m_TabSupport.prevSelectedControl[DATA_FIELD_SELECTION]:SetHide(true);
			end
			tabInst.Selection:SetHide(false);
			onClickCallback();
			CloseExtraTabs();
		end

		m_TabSupport.AddTab(tabInst.Button, callback);
	end

	return tabInst.Button;
end

function AddExtraTab(label:string, onClickCallback:ifunction)
	local extraTabInst:table = m_ExtraTabsIM:GetInstance();
	extraTabInst.Button:SetText(label);

	local callback = function()
		if(m_TabSupport.selectedControl ~= nil) then
			m_TabSupport.selectedControl[DATA_FIELD_SELECTION]:SetHide(true);
			m_TabSupport.SetSelectedTabVisually(nil);
		end
		for _,tabInst in pairs(m_ExtraTabs) do
			tabInst.Button:SetSelected(tabInst == extraTabInst);
		end
		onClickCallback();
	end

	extraTabInst.Button:RegisterCallback(Mouse.eLClick, callback);

	local textControl = extraTabInst.Button:GetTextControl();
	local textSize:number = textControl:GetSizeX();
	extraTabInst.Button:SetSizeX(textSize + PADDING_TAB_BUTTON_TEXT);
	extraTabInst.Button:RegisterCallback(Mouse.eMouseEnter,	function() UI.PlaySound("Main_Menu_Mouse_Over"); end);

	local tabSize:number = extraTabInst.Button:GetSizeX();
	if tabSize > m_MaxExtraTabSize then
		m_MaxExtraTabSize = tabSize;
	end

	table.insert(m_ExtraTabs, extraTabInst);
end

-- ===========================================================================
--	Called anytime player switches tabs
-- ===========================================================================
function ResetState(newView:ifunction)
	m_ActiveHeader = nil;
	m_ActiveViewUpdate = newView;
	
	Controls.OverallView:SetHide(true);
	Controls.ScoreView:SetHide(true);
	Controls.ScienceView:SetHide(true);
	Controls.CultureView:SetHide(true);
	Controls.DominationView:SetHide(true);
	Controls.ReligionView:SetHide(true);
	Controls.EconomicView:SetHide(true);
	Controls.MetropolisView:SetHide(true);
	Controls.TerritorialView:SetHide(true);
	Controls.DiplomaticView:SetHide(true);
	Controls.GenericView:SetHide(true);

	-- Reset tourism lens unless we're now view the Culture tab
	if newView ~= ViewCulture then
		ResetTourismLens();
	end
end

function ChangeActiveHeader(headerType:string, headerIM:table, parentControl:table)
	m_ActiveHeader = m_HeaderInstances[headerType];
	if(m_ActiveHeader == nil) then
		m_ActiveHeader = headerIM:GetInstance(parentControl);
		m_HeaderInstances[headerType] = m_ActiveHeader;
	end
end

function GetCivNameAndIcon(playerID:number, bColorUnmetPlayer:boolean)
	local name:string, icon:string;
	local playerConfig:table = PlayerConfigurations[playerID];
	
	if(playerID == m_LocalPlayerID or playerConfig:IsHuman() or m_LocalPlayer:GetDiplomacy():HasMet(playerID)) then
		name = Locale.Lookup(playerConfig:GetPlayerName());
		
		if playerID == m_LocalPlayerID or m_LocalPlayer:GetDiplomacy():HasMet(playerID) then
			icon = "ICON_" .. playerConfig:GetCivilizationTypeName();
		else
			icon = ICON_UNKNOWN_CIV;
		end
	else
		name = bColorUnmetPlayer and LOC_UNKNOWN_CIV_COLORED or LOC_UNKNOWN_CIV;
		icon = ICON_UNKNOWN_CIV;
	end
	
	return name, icon;
end

function ColorCivIcon(instance:table, playerID:number)
	if(playerID == m_LocalPlayerID or m_LocalPlayer:GetDiplomacy():HasMet(playerID)) then
		local backColor, frontColor = UI.GetPlayerColors(playerID);
		local darkerBackColor = DarkenLightenColor(backColor,(-55),230);
		local brighterBackColor = DarkenLightenColor(backColor,80,255);
		instance.CivIcon:SetColor(frontColor);
		if(instance.CivIconBacking ~= nil) then
			instance.CivIconBacking:SetColor(backColor);
		else
			instance.CivBacking_Base:SetColor(backColor);
			instance.CivBacking_Lighter:SetColor(brighterBackColor);
			instance.CivBacking_Darker:SetColor(darkerBackColor);
		end
	else
		instance.CivIcon:SetColor(UNKNOWN_COLOR);
		if(instance.CivIconBacking ~= nil) then
			instance.CivIconBacking:SetColor(UNKNOWN_COLOR);
		else
			instance.CivBacking_Base:SetColor(backColor);
			instance.CivBacking_Lighter:SetColor(brighterBackColor);
			instance.CivBacking_Darker:SetColor(darkerBackColor);
		end
	end
end

-- ===========================================================================
--	Called to update a generic header instance
-- ===========================================================================
function PopulateGenericHeader(resizeCallback:ifunction, title:string, subTitle:string, details:string, headerIcon:string, advisorIcon:string)
	
	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(headerIcon, SIZE_HEADER_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateGenericHeader: icon=\""..headerIcon.."\", iconSize="..tostring(SIZE_HEADER_ICON));
	else
		m_ActiveHeader.HeaderIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
	end

	m_ActiveHeader.HeaderLabel:SetText(Locale.ToUpper(Locale.Lookup(title)));
	if(subTitle ~= nil and subTitle ~= "") then
		m_ActiveHeader.HeaderSubLabel:SetHide(false);
		m_ActiveHeader.HeaderSubLabel:SetText(Locale.Lookup(subTitle));
	else
		m_ActiveHeader.HeaderSubLabel:SetHide(true);
	end

	m_ActiveHeader.AdvisorText:SetText(details and Locale.Lookup(details) or "");
	
	m_ActiveHeader.ExpandHeaderButton:RegisterCallback(Mouse.eLClick, OnExpandHeader);
	m_ActiveHeader.ContractHeaderButton:RegisterCallback(Mouse.eLClick, OnContractHeader);
	
	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED] == nil) then 
		m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED] = true;
	end

	m_ActiveHeader[DATA_FIELD_HEADER_RESIZED] = resizeCallback;
	RealizeHeaderSize();
end

-- ===========================================================================
--	Called anytime player presses expand/contract button on a Header Instance 
-- ===========================================================================
function OnExpandHeader(data1:number, data2:number, control:table)
	if(control == m_ActiveHeader.ExpandHeaderButton) then
		m_ActiveHeader.AdvisorIcon:SetHide(false);
		m_ActiveHeader.AdvisorText:SetHide(false);
		m_ActiveHeader.AdvisorTextBG:SetHide(false);
		m_ActiveHeader.ExpandHeaderButton:SetHide(true);
		m_ActiveHeader.ContractHeaderButton:SetHide(false);
		m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED] = true;
		
		RealizeHeaderSize();
	end
end
function OnContractHeader(data1:number, data2:number, control:table)
	if(control == m_ActiveHeader.ContractHeaderButton) then
		m_ActiveHeader.AdvisorIcon:SetHide(true);
		m_ActiveHeader.AdvisorText:SetHide(true);
		m_ActiveHeader.AdvisorTextBG:SetHide(true);
		m_ActiveHeader.ExpandHeaderButton:SetHide(false);
		m_ActiveHeader.ContractHeaderButton:SetHide(true);
		m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED] = false;
		
		RealizeHeaderSize();
	end
end

-- ===========================================================================
--	Called anytime header changes size (when it's expanded / contracted)
-- ===========================================================================
function RealizeHeaderSize()
	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local textBubbleHeight:number = m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_ADVISOR_TEXT_BG;
		
		if(textBubbleHeight > SIZE_HEADER_MAX_Y) then
			textBubbleHeight = SIZE_HEADER_MAX_Y;
		elseif textBubbleHeight < SIZE_HEADER_MIN_Y then
			textBubbleHeight = SIZE_HEADER_MIN_Y;
		end
		
		m_ActiveHeader.AdvisorTextBG:SetSizeY(textBubbleHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + textBubbleHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + textBubbleHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + textBubbleHeight);
		m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT] = textBubbleHeight;
	else
		m_ActiveHeader.HeaderFrame:SetSizeY(SIZE_HEADER_DEFAULT);
		m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT] = SIZE_HEADER_DEFAULT;
	end
	
	-- Center header label if sub label is not present
	if(m_ActiveHeader.HeaderSubLabel:IsHidden()) then
		m_ActiveHeader.HeaderLabel:SetOffsetY(25);
	else
		m_ActiveHeader.HeaderLabel:SetOffsetY(18);
	end
	
	if(m_ActiveHeader[DATA_FIELD_HEADER_RESIZED] ~= nil) then
		m_ActiveHeader[DATA_FIELD_HEADER_RESIZED]();
	end
end

-- ===========================================================================
--	Utility to reduce code duplication
-- ===========================================================================
function RealizeStackAndScrollbar(stackControl:table, scrollbarControl:table, offsetStackIfScrollbarHidden:boolean)
	stackControl:CalculateSize();
	stackControl:ReprocessAnchoring();
	scrollbarControl:CalculateInternalSize();
	scrollbarControl:ReprocessAnchoring();
	scrollbarControl:SetScrollValue(0);
	
	if(offsetStackIfScrollbarHidden ~= nil) then
		if(scrollbarControl:GetScrollBar():IsHidden()) then
			stackControl:SetOffsetX(OFFSET_HIDDEN_SCROLLBAR);
		else
			stackControl:SetOffsetX(0);
		end
	end
end

-- ===========================================================================
--	Called when Overall tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewOverall()
	ResetState(ViewOverall);
	Controls.OverallView:SetHide(false);

	m_OverallIM:ResetInstances();

	-- Add default victory types in a pre-determined order
	if(Game.IsVictoryEnabled("VICTORY_TECHNOLOGY")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_TECHNOLOGY", "SCIENCE");
	end
	if(Game.IsVictoryEnabled("VICTORY_CULTURE")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_CULTURE", "CULTURE");
	end
	if(Game.IsVictoryEnabled("VICTORY_CONQUEST")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_CONQUEST", "DOMINATION");
	end
	if(Game.IsVictoryEnabled("VICTORY_RELIGIOUS")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_RELIGIOUS", "RELIGION");
	end
	if(Game.IsVictoryEnabled("VICTORY_ECONOMIC")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_ECONOMIC", "ECONOMIC");
	end
	if (Game.IsVictoryEnabled("VICTORY_METROPOLIS")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_METROPOLIS", "METROPOLIS");
	end
	if (Game.IsVictoryEnabled("VICTORY_TERRITORIAL")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_TERRITORIAL", "TERRITORIAL");
	end
	if (Game.IsVictoryEnabled("VICTORY_DIPLOMATIC")) then
		PopulateOverallInstance(m_OverallIM:GetInstance(), "VICTORY_DIPLOMATIC", "DIPLOMATIC");
	end

	-- Add custom (modded) victory types
	for row in GameInfo.Victories() do
		local victoryType:string = row.VictoryType;
		if IsCustomVictoryType(victoryType) and Game.IsVictoryEnabled(victoryType) then
			PopulateOverallInstance(m_OverallIM:GetInstance(), victoryType);
		end
	end

	Controls.OverallViewStack:CalculateSize();
	Controls.OverallViewStack:ReprocessAnchoring();
	Controls.OverallViewScrollbar:CalculateInternalSize();
	Controls.OverallViewScrollbar:ReprocessAnchoring();

	if(Controls.OverallViewScrollbar:GetScrollBar():IsHidden()) then
		Controls.OverallViewStack:SetOffsetX(-3);
	else
		Controls.OverallViewStack:SetOffsetX(-5);
	end
end

function PopulateOverallInstance(instance:table, victoryType:string, typeText:string)
	
	local victoryInfo:table= GameInfo.Victories[victoryType];
	
	instance.VictoryLabel:SetText(Locale.ToUpper(Locale.Lookup(victoryInfo.Name)));
	instance.VictoryLabelUnderline:SetSizeX(instance.VictoryLabel:GetSizeX() + PADDING_VICTORY_LABEL_UNDERLINE);
	
	local icon:string;
	local color:number;
	
	if typeText ~= nil then
		icon = "ICON_VICTORY_" .. typeText;
		color = UI.GetColorValue("COLOR_VICTORY_" .. typeText);
	else
		icon = victoryInfo.Icon or ICON_GENERIC;
		color = UI.GetColorValue("White");
	end
	
	instance.VictoryBanner:SetColor(color);
	instance.VictoryLabelGradient:SetColor(color);

	if icon ~= nil then
		local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(icon, SIZE_VICTORY_ICON_SMALL);
		
		if(textureSheet == nil or textureSheet == "") then
			UI.DataError("Could not find icon in PopulateOverallInstance: icon=\""..icon.."\", iconSize="..tostring(SIZE_VICTORY_ICON_SMALL));
		else
			instance.VictoryIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
			instance.VictoryIcon:SetHide(false);
		end
	else
		instance.VictoryIcon:SetHide(true);
	end

	-- Tiebreak score functions
	local firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_DOMINATION_SCORE";
	local firstTiebreakerFunction = function(p)
		return p:GetScore();
	end;
	local secondTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_DOMINATION_SCORE";
	local secondTiebreakerFunction = function(p)
		return p:GetScore();
	end;
	
	if (victoryType == "VICTORY_TECHNOLOGY") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_SCIENCE_NUM_TECHS";
		firstTiebreakerFunction = function(p)
			return p:GetStats():GetNumTechsResearched();
		end;
		secondTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_SCIENCE_SCIENCE_RATE";
		secondTiebreakerFunction = function(p)
			return p:GetTechs():GetScienceYield();
		end;
	elseif (victoryType == "VICTORY_CULTURE") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_CULTURE_TOURISM_RATE";
		firstTiebreakerFunction = function(p)
			return p:GetStats():GetTourism();
		end;
		secondTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_CULTURE_CULTURE_RATE";
		secondTiebreakerFunction = function(p)
			return p:GetCulture():GetCultureYield();
		end;
	elseif (victoryType == "VICTORY_CONQUEST") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_DOMINATION_MILITARY_STRENGTH";
		firstTiebreakerFunction = function(p)
			return p:GetStats():GetMilitaryStrength();
		end;
	elseif (victoryType == "VICTORY_RELIGIOUS") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_RELIGION_CITIES_FOLLOWING_RELIGION";
		firstTiebreakerFunction = function(p)
			return p:GetStats():GetNumCitiesFollowingReligion();
		end;
		secondTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_RELIGION_FAITH_RATE";
		secondTiebreakerFunction = function(p)
			return p:GetReligion():GetFaithYield();
		end;
	
	-- // -----------------------------------------------------------------------------
	-- // CiVI Reformation Victories
	-- // -----------------------------------------------------------------------------
	elseif (victoryType == "VICTORY_ECONOMIC") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_ECONOMIC_GOLD";
		firstTiebreakerFunction = function(p)
			return GetEconomicVictoryGoldPerTurn(p);
		end;
		-- secondTiebreakerText already defined at start of function
	elseif (victoryType == "VICTORY_METROPOLIS") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_METROPOLIS_OVERVIEW_LARGEST_CITY";
		firstTiebreakerFunction = function(p)
			return CalculateMetropolisLargestCity(p);
		end;
		-- secondTiebreakerText already defined at start of function
	elseif (victoryType == "VICTORY_TERRITORIAL") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_TERRITORIAL_OVERVIEW_TERRITORY";
		firstTiebreakerFunction = function(p)
			return GetTerritorialVictoryTerritorySizeAsPercentage(p);
		end;
		secondTiebreakerText = "LOC_WORLD_RANKINGS_OVERVIEW_DOMINATION_MILITARY_STRENGTH";
		secondTiebreakerFunction = function(p)
			return p:GetStats():GetMilitaryStrength();
		end;
	elseif (victoryType == "VICTORY_DIPLOMATIC") then
		firstTiebreakerText = "LOC_WORLD_RANKINGS_DIPLOMATIC_OVERVIEW_DIPLOMATIC_POWER";
		firstTiebreakerFunction = function(p)
			return GetDiplomaticVictoryDiplomaticPower(p);
		end;
		-- secondTiebreakerText already defined at start of function
	end
	
	if (victoryType == "VICTORY_ECONOMIC") then
		SetVictoryBannerInfo(instance, "Victory_Economic.dds", UI.GetColorValue("COLOR_VICTORY_ECONOMIC"));
	end
	if (victoryType == "VICTORY_METROPOLIS") then
		SetVictoryBannerInfo(instance, "Victory_Metropolis.dds", UI.GetColorValue("COLOR_VICTORY_METROPOLIS"));
	end
	if (victoryType == "VICTORY_TERRITORIAL") then
		SetVictoryBannerInfo(instance, "Victory_Territorial.dds", UI.GetColorValue("COLOR_VICTORY_TERRITORIAL"));
	end
	if (victoryType == "VICTORY_DIPLOMATIC") then
		SetVictoryBannerInfo(instance, "Victory_Diplomatic.dds", UI.GetColorValue("COLOR_VICTORY_DIPLOMATIC"));
	end

	local numPlayers:number = 0;
	local playersData:table = {};
	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			table.insert(playersData, {
				Player = pPlayer,
				--Score = Game.GetVictoryProgressForPlayer(victoryType, i), -- Game Core Call
				Score = GetCustomVictoryProgressForPlayer(victoryType, pPlayer, i),
				FirstTiebreakScore = firstTiebreakerFunction(pPlayer),
				SecondTiebreakScore = secondTiebreakerFunction(pPlayer),
				FirstTiebreakSummary = Locale.Lookup(firstTiebreakerText, Round(firstTiebreakerFunction(pPlayer), 1)),
				SecondTiebreakSummary = Locale.Lookup(secondTiebreakerText, Round(secondTiebreakerFunction(pPlayer), 1)),
			});
			numPlayers = numPlayers + 1;
		end
	end

	-- Ensure we have Instance Managers for the player meters
	local playersIM:table = instance[DATA_FIELD_OVERALL_PLAYERS_IM];
	if(playersIM == nil) then
		playersIM = InstanceManager:new("OverallPlayerInstance", "CivIconBackingFaded", instance.PlayerStack);
		instance[DATA_FIELD_OVERALL_PLAYERS_IM] = playersIM;
	end
	playersIM:ResetInstances();
	
	if(numPlayers > 0) then
		-- Sort players by Score, including tiebreakers
		table.sort(playersData, function(a, b)
			if (a.Score == b.Score) then
				if (a.FirstTiebreakScore == b.FirstTiebreakScore) then
					if (a.SecondTiebreakScore == b.SecondTiebreakScore) then
						return a.Player:GetID() < b.Player:GetID();
					end
					return a.SecondTiebreakScore > b.SecondTiebreakScore;
				end
				return a.FirstTiebreakScore > b.FirstTiebreakScore;
			end
			return a.Score > b.Score;
		end);

		-- Populate top player icon
		PopulateOverallPlayerIconInstance(instance, playersData[1], SIZE_OVERALL_TOP_PLAYER_ICON);

		-- Populate leading and local player labels
		local topPlayer:number = playersData[1].Player:GetID();
		if(topPlayer == m_LocalPlayerID) then
			instance.VictoryPlayer:SetText("");
			instance.VictoryLeading:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_FIRST_PLACE_YOU_SIMPLE"));
		else
			-- Determine local player position
			local localPlayerPosition:number = 1;
			local localPlayerPositionText:string;
			local localPlayerPositionLocTag:string;
			for i = 1, numPlayers do
				if(playersData[i].Player == m_LocalPlayer) then
					localPlayerPosition = i;
					break;
				end
			end

			-- Generate position text (ex: "You are in third place" or "You are in position 15")
			if(localPlayerPosition <= 12) then
				localPlayerPositionText = Locale.Lookup("LOC_WORLD_RANKINGS_" .. localPlayerPosition .. "_PLACE");
				localPlayerPositionLocTag = "LOC_WORLD_RANKINGS_OTHER_PLACE_";
			else
				localPlayerPositionText = tostring(localPlayerPosition);
				localPlayerPositionLocTag = "LOC_WORLD_RANKINGS_OTHER_POSITION_";
			end
			localPlayerPositionLocTag = localPlayerPositionLocTag .. "SIMPLE";
			instance.VictoryPlayer:SetText(Locale.Lookup(localPlayerPositionLocTag, localPlayerPositionText));

			-- Set top player position text
			local topCivName:string;
			if(m_LocalPlayer:GetDiplomacy():HasMet(topPlayer)) then
				topCivName = Locale.Lookup(GameInfo.Civilizations[PlayerConfigurations[topPlayer]:GetCivilizationTypeID()].Name);
			else
				topCivName = LOC_UNKNOWN_CIV;
			end
			instance.VictoryLeading:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_FIRST_PLACE_OTHER_SIMPLE", topCivName));
		end
		
		-- Spawn other player icons
		if(numPlayers > 1) then
			for i = 2, numPlayers do
				local playerInstance:table = playersIM:GetInstance();
				PopulateOverallPlayerIconInstance(playerInstance, playersData[i], SIZE_OVERALL_PLAYER_ICON);
			end
		end

		instance.ButtonBG:SetSizeY(SIZE_OVERALL_BG_HEIGHT + math.max(instance.PlayerStack:GetSizeY(), SIZE_OVERALL_INSTANCE));
	end
end

function PopulateOverallPlayerIconInstance(instance:table, playerData:table, iconSize:number)
	local playerID:number = playerData.Player:GetID();

	ColorCivIcon(instance, playerID);

	local civName:string, civIcon:string = GetCivNameAndIcon(playerID);
	local details:string = playerData.FirstTiebreakSummary .. "[NEWLINE]" .. playerData.SecondTiebreakSummary;
	local textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, iconSize);
	
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateOverallPlayerIconInstance: icon=\""..civIcon.."\", iconSize="..tostring(iconSize));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		instance.CivIconFaded:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		instance.CivIcon:SetPercent(playerData.Score);
		instance.CivIconBacking:SetPercent(playerData.Score);
		SetLeaderTooltip(instance, playerID, details);
	end

	instance.LocalPlayer:SetHide(playerID ~= m_LocalPlayerID);
end

function SetLeaderTooltip(instance:table, playerID:number, details:string)
	local pPlayer:table = Players[playerID];
	local playerConfig:table = PlayerConfigurations[playerID];
	if(playerID ~= m_LocalPlayerID and not m_LocalPlayer:GetDiplomacy():HasMet(playerID)) then
		instance.CivIcon:SetToolTipType();
		instance.CivIcon:ClearToolTipCallback();
		if GameConfiguration.IsAnyMultiplayer() and pPlayer:IsHuman() then
			instance.CivIcon:SetToolTipString(Locale.Lookup("LOC_DIPLOPANEL_UNMET_PLAYER") .. " (" .. playerConfig:GetPlayerName() .. ")");
		else
			instance.CivIcon:SetToolTipString(Locale.Lookup("LOC_DIPLOPANEL_UNMET_PLAYER"));
		end
	else
		instance.CivIcon:SetToolTipType("CivTooltip");
		instance.CivIcon:SetToolTipCallback(function() UpdateLeaderTooltip(instance, playerID, details); end);
	end
end

function UpdateLeaderTooltip(instance:table, playerID:number, details:string)
	local pPlayer:table = Players[playerID];
	local playerConfig:table = PlayerConfigurations[playerID];
	if(pPlayer ~= nil and playerConfig ~= nil) then

		m_CivTooltip.YouIndicator:SetHide(playerID ~= m_LocalPlayerID);

		local leaderTypeName:string = playerConfig:GetLeaderTypeName();
		if(leaderTypeName ~= nil) then
			local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas("ICON_"..leaderTypeName, SIZE_LEADER_ICON);
			if(textureSheet == nil or textureSheet == "") then
				UI.DataError("Could not find icon in UpdateLeaderTooltip: icon=\"".."ICON_"..leaderTypeName.."\", iconSize="..tostring(SIZE_LEADER_ICON));
			else
				m_CivTooltip.LeaderIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
			end

			local desc:string;
			local leaderDesc:string = playerConfig:GetLeaderName();
			local civDesc:string = playerConfig:GetCivilizationDescription();
			if GameConfiguration.IsAnyMultiplayer() and pPlayer:IsHuman() then
				desc = Locale.Lookup("LOC_DIPLOMACY_DEAL_PLAYER_PANEL_TITLE", leaderDesc, civDesc) .. " (" .. playerConfig:GetPlayerName() .. ")";
			else
				desc = Locale.Lookup("LOC_DIPLOMACY_DEAL_PLAYER_PANEL_TITLE", leaderDesc, civDesc);
			end
			
			if (details ~= nil and details ~= "") then
				desc = desc .. "[NEWLINE]" .. details;
			end
			m_CivTooltip.LeaderName:SetText(desc);
			m_CivTooltip.BG:DoAutoSize();
		end
	end
end

-- ===========================================================================
--	Called when Score tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewScore()
	ResetState(ViewScore);
	Controls.ScoreView:SetHide(false);

	ChangeActiveHeader("VICTORY_SCORE", m_GenericHeaderIM, Controls.ScoreViewHeader);
	local subTitle:string = Locale.Lookup("LOC_WORLD_RANKINGS_SCORE_CONDITION", Game.GetMaxGameTurns());
	PopulateGenericHeader(RealizeScoreStackSize, SCORE_TITLE, subTitle, SCORE_DETAILS, ICON_GENERIC);

	m_ScoreIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateScoreInstance(m_ScoreIM:GetInstance(), pPlayer);
		end
	end

	RealizeScoreStackSize();
end

function PopulateScoreInstance(instance:table, pPlayer:table)
	local playerID:number = pPlayer:GetID();
	local civName, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateScoreInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	instance.CivName:SetText(civName);
	instance.Score:SetText(pPlayer:GetScore());
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	if(m_ShowScoreDetails) then
		instance.ButtonBG:SetSizeY(SIZE_SCORE_ITEM_DETAILS);
		
		local detailsText:string = "";
		local scoreCategories = GameInfo.ScoringCategories;
		local numCategories:number = #scoreCategories;
		for i = 0, numCategories - 1 do
			if scoreCategories[i].Multiplier > 0 or m_isDebugForceShowAllScoreCategories then
				local category:table = scoreCategories[i];
				detailsText = detailsText .. Locale.Lookup(category.Name) .. ": " .. pPlayer:GetCategoryScore(i);
				if(i <= numCategories) then
					detailsText = detailsText .. "[NEWLINE]";
				end
			end
		end
		instance.Details:SetText(detailsText);
		instance.Details:SetHide(false);
	else
		instance.ButtonBG:SetSizeY(SIZE_SCORE_ITEM_DEFAULT);
		instance.Details:SetHide(true);
	end
end

function RealizeScoreStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		Controls.ScoreViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.ScoreViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
	else
		Controls.ScoreViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.ScoreViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
	end

	RealizeStackAndScrollbar(Controls.ScoreViewStack, Controls.ScoreViewScrollbar, true);

	local textSize:number = Controls.ScoreDetailsButton:GetTextControl():GetSizeX();
	Controls.ScoreDetailsButton:SetSizeX(textSize + PADDING_SCORE_DETAILS_BUTTON_WIDTH);
end

function ToggleScoreDetails()
	m_ShowScoreDetails = not m_ShowScoreDetails;
	Controls.ScoreDetailsCheck:SetCheck(m_ShowScoreDetails);
	ViewScore();
end

-- ===========================================================================
--	Called when Science tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewScience()
	ResetState(ViewScience);
	Controls.ScienceView:SetHide(false);

	ChangeActiveHeader("VICTORY_TECHNOLOGY", m_ScienceHeaderIM, Controls.ScienceViewHeader);
	PopulateGenericHeader(RealizeScienceStackSize, SCIENCE_TITLE, "", SCIENCE_DETAILS, SCIENCE_ICON);
	
	local totalCost:number = 0;
	local currentProgress:number = 0;
	local progressText:string = "";
	local progressResults:table = { 0, 0, 0 }; -- initialize with 3 elements
	local finishedProjects:table = { {}, {}, {} };
	
	local bHasSpaceport:boolean = false;
	for _,district in m_LocalPlayer:GetDistricts():Members() do
		if (district ~= nil and district:IsComplete() and district:GetType() == SPACE_PORT_DISTRICT_INFO.Index) then
			bHasSpaceport = true;
			break;
		end
	end

	local pPlayerStats:table = m_LocalPlayer:GetStats();
	local pPlayerCities:table = m_LocalPlayer:GetCities();
	for _, city in pPlayerCities:Members() do
		local pBuildQueue:table = city:GetBuildQueue();
		-- 1st milestone - satelite launch
		totalCost = 0;
		currentProgress = 0;
		for i, projectInfo in ipairs(EARTH_SATELLITE_PROJECT_INFOS) do
			local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
			local projectProgress:number = projectCost;
			if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
				projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
			end
			totalCost = totalCost + projectCost;
			currentProgress = currentProgress + projectProgress;
			finishedProjects[1][i] = projectProgress == projectCost;
		end
		progressResults[1] = currentProgress / totalCost;

		-- 2nd milestone - moon landing
		totalCost = 0;
		currentProgress = 0;
		for i, projectInfo in ipairs(MOON_LANDING_PROJECT_INFOS) do
			local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
			local projectProgress:number = projectCost;
			if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
				projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
			end
			totalCost = totalCost + projectCost;
			currentProgress = currentProgress + projectProgress;
			finishedProjects[2][i] = projectProgress == projectCost;
		end
		progressResults[2] = currentProgress / totalCost;
		
		-- 3rd milestone - mars landing
		totalCost = 0;
		currentProgress = 0;
		for i, projectInfo in ipairs(MARS_COLONY_PROJECT_INFOS) do
			local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
			local projectProgress:number = projectCost;
			if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
				projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
			end
			totalCost = totalCost + projectCost;
			currentProgress = currentProgress + projectProgress;
			finishedProjects[3][i] = projectProgress == projectCost;
		end
		progressResults[3] = currentProgress / totalCost;
	end

	local nextStep:string = "";
	for i, result in ipairs(progressResults) do
		if(result < 1) then
			progressText = progressText .. "[ICON_Bolt]";
			if(nextStep == "") then
				nextStep = GetNextStepForScienceProject(m_LocalPlayer, SCIENCE_PROJECTS[i], bHasSpaceport, finishedProjects[i]);
			end
		else
			progressText = progressText .. "[ICON_CheckmarkBlue] ";
		end
		progressText = progressText .. SCIENCE_REQUIREMENTS[i];
		if(i < 3) then progressText = progressText .. "[NEWLINE]"; end
	end

	m_ActiveHeader.AdvisorTextCentered:SetText(progressText);
	m_ActiveHeader.AdvisorTextNextStep:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_NEXT_STEP", nextStep));

	m_ScienceIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateScienceInstance(m_ScienceIM:GetInstance(), pPlayer);
		end
	end

	RealizeScienceStackSize();
end

function GetNextStepForScienceProject(pPlayer:table, projectInfos:table, bHasSpaceport:boolean, finishedProjects:table)

	if(not bHasSpaceport) then 
		return Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_NEXT_STEP_BUILD", Locale.Lookup(SPACE_PORT_DISTRICT_INFO.Name));
	end

	local playerTech:table = pPlayer:GetTechs();
	local numProjectInfos:number = table.count(projectInfos);
	for i, projectInfo in ipairs(projectInfos) do

		if(projectInfo.PrereqTech ~= nil) then
			local tech:table = GameInfo.Technologies[projectInfo.PrereqTech];
			if(not playerTech:HasTech(tech.Index)) then
				return Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_NEXT_STEP_RESEARCH", Locale.Lookup(tech.Name));
			end
		end

		if(not finishedProjects[i]) then
			return Locale.Lookup(projectInfo.Name);
		end
	end
	return "";
end

function PopulateScienceInstance(instance:table, pPlayer:table)
	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateScoreInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	TruncateStringWithTooltip(instance.CivName, 180, civName); 
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local bHasMet = m_LocalPlayer:GetDiplomacy():HasMet(playerID);
	if (bHasMet == true or playerID == Game.GetLocalPlayer()) then

		local bHasSpaceport:boolean = false;
		for _,district in pPlayer:GetDistricts():Members() do
			if (district ~= nil and district:IsComplete() and district:GetType() == SPACE_PORT_DISTRICT_INFO.Index) then
				bHasSpaceport = true;
				break;
			end
		end

		local pPlayerStats:table = pPlayer:GetStats();
		local pPlayerCities:table = pPlayer:GetCities();
		local projectTotals:table = { 0, 0, 0 };
		local projectProgresses:table = { 0, 0, 0 };
		local finishedProjects:table = { {}, {}, {} };
		for _, city in pPlayerCities:Members() do
			local pBuildQueue:table = city:GetBuildQueue();

			-- 1st milestone - satelite launch
			for i, projectInfo in ipairs(EARTH_SATELLITE_PROJECT_INFOS) do
				local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
				local projectProgress:number = projectCost;
				if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
					projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
				end
				finishedProjects[1][i] = false;
				if projectProgress ~= 0 then
					projectTotals[1] = projectTotals[1] + projectCost;
					projectProgresses[1] = projectProgresses[1] + projectProgress;
					finishedProjects[1][i] = projectProgress == projectCost;
				end
			end

			-- 2nd milestone - moon landing
			for i, projectInfo in ipairs(MOON_LANDING_PROJECT_INFOS) do
				local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
				local projectProgress:number = projectCost;
				if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
					projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
				end
				finishedProjects[2][i] = false;
				if projectProgress ~= 0 then
					projectTotals[2] = projectTotals[2] + projectCost;
					projectProgresses[2] = projectProgresses[2] + projectProgress;
					finishedProjects[2][i] = projectProgress == projectCost;
				end
			end

			-- 3rd milestone - mars landing
			for i, projectInfo in ipairs(MARS_COLONY_PROJECT_INFOS) do
				local projectCost:number = pBuildQueue:GetProjectCost(projectInfo.Index);
				local projectProgress:number = projectCost;
				if pPlayerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0 then
					projectProgress = pBuildQueue:GetProjectProgress(projectInfo.Index);
				end
				finishedProjects[3][i] = false;
				if projectProgress ~= 0 then
					projectTotals[3] = projectTotals[3] + projectCost;
					projectProgresses[3] = projectProgresses[3] + projectProgress;
					finishedProjects[3][i] = projectProgress == projectCost;
				end
			end
		end

		instance.ObjFill_1:SetHide(projectProgresses[1] == 0);
		instance.ObjBar_1:SetPercent(projectProgresses[1] / projectTotals[1]);
		instance.ObjToggle_ON_1:SetHide(projectTotals[1] == 0 or projectProgresses[1] ~= projectTotals[1]);

		instance.ObjFill_2:SetHide(projectProgresses[2] == 0);
		instance.ObjBar_2:SetPercent(projectProgresses[2] / projectTotals[2]);
		instance.ObjToggle_ON_2:SetHide(projectTotals[2] == 0 or projectProgresses[2] ~= projectTotals[2]);

		instance.ObjFill_3:SetHide(projectProgresses[3] == 0);
		instance.ObjBar_3:SetPercent(projectProgresses[3] / projectTotals[3]);
		instance.ObjToggle_ON_3:SetHide(projectTotals[3] == 0 or projectProgresses[3] ~= projectTotals[3]);
		
		instance.ObjBG_1:SetToolTipString(GetTooltipForScienceProject(pPlayer, EARTH_SATELLITE_PROJECT_INFOS, bHasSpaceport, finishedProjects[1]));
		instance.ObjBG_2:SetToolTipString(GetTooltipForScienceProject(pPlayer, MOON_LANDING_PROJECT_INFOS, nil, finishedProjects[2]));
		instance.ObjBG_3:SetToolTipString(GetTooltipForScienceProject(pPlayer, MARS_COLONY_PROJECT_INFOS, nil, finishedProjects[3]));

		for i = 1, 3 do
			instance["ObjHidden_" .. i]:SetHide(true);
		end
	else -- Unmet civ
		for i = 1, 3 do
			instance["ObjBar_" .. i]:SetPercent(0);
			instance["ObjFill_" .. i]:SetHide(true);
			instance["ObjHidden_" .. i]:SetHide(false);
			instance["ObjToggle_ON_" .. i]:SetHide(true);
			instance["ObjToggle_ON_" .. i]:SetToolTipString("");
			instance["ObjToggle_OFF_" .. i]:SetToolTipString("");
		end
	end
end

function GetTooltipForScienceProject(pPlayer:table, projectInfos:table, bHasSpaceport:boolean, finishedProjects:table)

	local result:string = "";

	-- Only show spaceport for first tooltip
	if bHasSpaceport ~= nil then
		if(bHasSpaceport) then 
			result = result .. "[ICON_CheckmarkBlue]";
		else
			result = result .. "[ICON_Bolt]";
		end
		result = result .. Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_NEXT_STEP_BUILD", Locale.Lookup(SPACE_PORT_DISTRICT_INFO.Name)) .. "[NEWLINE]";
	end

	local playerTech:table = pPlayer:GetTechs();
	local numProjectInfos:number = table.count(projectInfos);
	for i, projectInfo in ipairs(projectInfos) do

		if(projectInfo.PrereqTech ~= nil) then
			local tech:table = GameInfo.Technologies[projectInfo.PrereqTech];
			if(playerTech:HasTech(tech.Index)) then
				result = result .. "[ICON_CheckmarkBlue]";
			else
				result = result .. "[ICON_Bolt]";
			end
			result = result .. Locale.Lookup("LOC_WORLD_RANKINGS_SCIENCE_NEXT_STEP_RESEARCH", Locale.Lookup(tech.Name)) .. "[NEWLINE]";
		end

		if(finishedProjects[i]) then
			result = result .. "[ICON_CheckmarkBlue]";
		else
			result = result .. "[ICON_Bolt]";
		end
		result = result .. Locale.Lookup(projectInfo.Name);
		if(i < numProjectInfos) then result = result .. "[NEWLINE]"; end
	end

	return result;
end

function RealizeScienceStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		m_ActiveHeader.AdvisorTextCentered:SetOffsetY(OFFSET_SCIENCE_REQUIREMENTS_Y + m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_HEADER);
		m_ActiveHeader.AdvisorTextNextStep:SetOffsetY(m_ActiveHeader.AdvisorTextCentered:GetOffsetY() + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + PADDING_HEADER);
		m_ActiveHeader.NextStepHighlight:SetOffsetY(-m_ActiveHeader.AdvisorTextNextStep:GetSizeY() + PADDING_NEXT_STEP_HIGHLIGHT);
		headerHeight = headerHeight + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + m_ActiveHeader.AdvisorTextNextStep:GetSizeY() + (PADDING_HEADER * 2);
		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);
		Controls.ScienceViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.ScienceViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
		m_ActiveHeader.AdvisorTextCentered:SetHide(false);
		m_ActiveHeader.AdvisorTextNextStep:SetHide(false);
	else
		Controls.ScienceViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.ScienceViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		m_ActiveHeader.AdvisorTextCentered:SetHide(true);
		m_ActiveHeader.AdvisorTextNextStep:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.ScienceViewStack, Controls.ScienceViewScrollbar, true);
end

-- ===========================================================================
function ResetTourismLens()
	if UILens.IsLensActive("TourismWithUnits") then
		UILens.SetActive("Default");
	end
end

-- ===========================================================================
--	Called when Culture tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function SetCultureHeaderState(state:number)
	m_CultureHeaderState = state;
	ViewCulture();
end

function ViewCulture()
	ResetState(ViewCulture);
	Controls.CultureView:SetHide(false);

	ChangeActiveHeader("VICTORY_CULTURE", m_CultureHeaderIM, Controls.CultureViewHeader);

	-- Active the Tourism lens if it isn't already
	if not UILens.IsLensActive("TourismWithUnits") then
		UILens.SetActive("TourismWithUnits");
	end

	local detailsText:string;
	if(m_CultureHeaderState == CULTURE_HEADER_STATES.WHAT_IS_CULTURE_VICTORY) then
		detailsText = CULTURE_VICTORY_DETAILS;
		m_ActiveHeader.DomesticTourism:SetText(CULTURE_DOMESTIC_TOURISTS);
		m_ActiveHeader.VisitingTourism:SetText(CULTURE_VISITING_TOURISTS);
	else
		UI.DataError("Unknown m_CultureHeaderState in ViewCulture: " .. tostring(m_CultureHeaderState));
	end

	PopulateGenericHeader(RealizeCultureStackSize, CULTURE_TITLE, "", detailsText, CULTURE_ICON);

	-- Pre-populate compute number of tourists required for victory
	local iTouristsNeededForWin = {};
	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			if(iTouristsNeededForWin[pPlayer] == nil) then
				iTouristsNeededForWin[pPlayer] = 0;
			end
			local iStaycationers = pPlayer:GetCulture():GetStaycationers();
			-- Raise numbers needed for all other players if higher
			for j = 0, PlayerManager.GetWasEverAliveCount() - 1 do
				local pTargetPlayer = Players[j];
				if (i ~= j and pTargetPlayer:IsAlive() == true and pTargetPlayer:IsMajor() == true) then
					if(iTouristsNeededForWin[pTargetPlayer] == nil) then
						iTouristsNeededForWin[pTargetPlayer] = 0;
					end
					if (iStaycationers >= iTouristsNeededForWin[pTargetPlayer]) then
						iTouristsNeededForWin[pTargetPlayer] = iStaycationers + 1;
					end
				end
			end
		end
	end
	table.sort(iTouristsNeededForWin, function(a, b) return a < b; end);

	m_CultureIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateCultureInstance(m_CultureIM:GetInstance(), pPlayer, iTouristsNeededForWin[pPlayer]);
		end
	end

	RealizeCultureStackSize();
end

function PopulateCultureInstance(instance:table, pPlayer:table, iTouristsNeededForWin:number)
	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateScoreInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		local tourism:string = Locale.Lookup("LOC_WORLD_RANKINGS_OVERVIEW_CULTURE_TOURISM_RATE", Round(pPlayer:GetStats():GetTourism(), 1));
		local culture:string = Locale.Lookup("LOC_WORLD_RANKINGS_OVERVIEW_CULTURE_CULTURE_RATE", Round(pPlayer:GetCulture():GetCultureYield(), 1));
		local details:string = tourism .. "[NEWLINE]" .. culture;

		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID, details);
	end
	
	ColorCivIcon(instance, playerID);
	instance.CivName:LocalizeAndSetText(civName);
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local numTourists:number = pPlayer:GetCulture():GetTouristsTo();
	instance.VisitingTourists:SetText(numTourists .. "/" .. iTouristsNeededForWin);
	instance.TouristsFill:SetPercent(numTourists / iTouristsNeededForWin);
	if(playerID == m_LocalPlayerID) then
		instance.VisitingUsContainer:SetHide(true);
	else
		instance.VisitingUsContainer:SetHide(false);
	end
	local backColor, _ = UI.GetPlayerColors(playerID);
	local brighterBackColor = DarkenLightenColor(backColor,35,255);
	if(playerID == m_LocalPlayerID or m_LocalPlayer:GetDiplomacy():HasMet(playerID)) then
		instance.DomesticTouristsIcon:SetColor(brighterBackColor);
	else
		instance.DomesticTouristsIcon:SetColor(RGBAValuesToABGRHex(1, 1, 1, 0.35));
	end
	instance.DomesticTourists:SetText(pPlayer:GetCulture():GetStaycationers());

	instance.VisitingUsTourists:SetText(m_LocalPlayer:GetCulture():GetTouristsFrom(playerID));
	instance.VisitingUsTourists:SetToolTipString(m_LocalPlayer:GetCulture():GetTouristsFromTooltip(playerID));
	instance.VisitingUsIcon:SetToolTipString(m_LocalPlayer:GetCulture():GetTouristsFromTooltip(playerID));
end

function RealizeCultureStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_ADVISOR_TEXT_BG;

		if(m_CultureHeaderState == CULTURE_HEADER_STATES.WHAT_IS_CULTURE_VICTORY) then
			m_ActiveHeader.HowToAttractTourists:SetOffsetY(headerHeight);
			headerHeight = headerHeight + math.max(m_ActiveHeader.DomesticTourism:GetSizeY(), m_ActiveHeader.DomesticTourismIcon:GetSizeY()) + (PADDING_HEADER * 2);
			headerHeight = headerHeight + math.max(m_ActiveHeader.VisitingTourism:GetSizeY(), m_ActiveHeader.VisitingTourismIcon:GetSizeY()) + PADDING_HEADER;
			m_ActiveHeader.HowToAttractTourists:SetHide(false);
		else
			UI.DataError("Unknown m_CultureHeaderState in ViewCulture: " .. tostring(m_CultureHeaderState));
		end

		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);

		m_ActiveHeader.StateBG:SetOffsetY(headerHeight + PADDING_CULTURE_HEADER);

		headerHeight = headerHeight + m_ActiveHeader.StateBG:GetSizeY() + PADDING_HEADER;

		m_ActiveHeader.NextState:SetOffsetX(m_ActiveHeader.StateBG:GetSizeX());
		
		Controls.CultureViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.CultureViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
	else
		Controls.CultureViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.CultureViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		m_ActiveHeader.HowToAttractTourists:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.CultureViewStack, Controls.CultureViewScrollbar, true);
end

-- ===========================================================================
--	Called when Domination tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewDomination()
	ResetState(ViewDomination);
	Controls.DominationView:SetHide(false);

	ChangeActiveHeader("VICTORY_CONQUEST", m_GenericHeaderIM, Controls.DominationViewHeader);
	PopulateGenericHeader(RealizeDominationStackSize, DOMINATION_TITLE, "", DOMINATION_DETAILS, DOMINATION_ICON);

	-- Pre-populate captured capitals
	local dominationData:table = {};

	for i, playerId in ipairs(PlayerManager.GetWasEverAliveMajorIDs()) do
		local pPlayer = Players[playerId];

		if (pPlayer:IsAlive()) then
			local pCities = pPlayer:GetCities();

			-- Don't show player if they haven't founded a capital city yet
			local pCapital = pCities:GetCapitalCity();
			if(pCapital ~= nil) then
			
				local data = {};
				for _, city in pCities:Members() do
					local originalOwnerID:number = city:GetOriginalOwner();
					local pOriginalOwner:table = Players[originalOwnerID];
					if(playerId ~= originalOwnerID and pOriginalOwner:IsMajor() and city:IsOriginalCapital()) then
						table.insert(data, originalOwnerID);
					end
				end
				table.insert(dominationData, {playerId, data});
			end
		end
	end

	m_DominationIM:ResetInstances();

	if (#dominationData > 0) then
		table.sort(dominationData, function(a, b) return #a[2] > #b[2] end);
	end

	for i, v in ipairs(dominationData) do
		local pPlayer = Players[v[1]];
		local dominatedCities = v[2];
		PopulateDominationInstance(m_DominationIM:GetInstance(), pPlayer, dominatedCities);
	end

	RealizeDominationStackSize();
end

function PopulateDominationInstance(instance:table, pPlayer:table, dominatedCities:table)
	local playerID:number = pPlayer:GetID();
	local backColor, frontColor = UI.GetPlayerColors(playerID);
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateDominationInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end
	
	ColorCivIcon(instance, playerID);
	instance.CivName:SetText(civName);
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local pCities = pPlayer:GetCities();
	local pCapital = pCities:GetCapitalCity();

	local hasOriginalCapital = false;
	for _, city in pCities:Members() do
		if(city:IsOriginalCapital() and city:GetOriginalOwner() == playerID) then
			hasOriginalCapital = true;
			break;
		end
	end

	instance.HasCapital:SetHide(not hasOriginalCapital);
	instance.HasCapital:SetToolTipString(DOMINATION_HAS_ORIGINAL_CAPITAL);

	local dominatedCitiesIM:table = instance[DATA_FIELD_DOMINATED_CITIES_IM];
	if(dominatedCitiesIM == nil) then
		dominatedCitiesIM = InstanceManager:new("DominatedCapitalInstance", "CivIconBacking", instance.CapitalsCapturedStack);
		instance[DATA_FIELD_DOMINATED_CITIES_IM] = dominatedCitiesIM;
	end
	dominatedCitiesIM:ResetInstances();

	local numDominatedCities:number = 0;
	if(dominatedCities ~= nil) then
		for _, dominatedPlayerID in pairs(dominatedCities) do
			numDominatedCities = numDominatedCities + 1;

			local dominatedInstance = dominatedCitiesIM:GetInstance();
			ColorCivIcon(dominatedInstance, dominatedPlayerID);
			
			local civName:string, civIcon:string = GetCivNameAndIcon(dominatedPlayerID, true);
			textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, SIZE_RELIGION_CIV_ICON);
			if(textureSheet == nil or textureSheet == "") then
				UI.DataError("Could not find icon for dominated civ in PopulateDominationInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_RELIGION_CIV_ICON));
			else
				dominatedInstance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
				SetLeaderTooltip(dominatedInstance, dominatedPlayerID);
			end
		end
	end

	instance.CapitalsCaptured:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_DOMINATION_SUMMARY", numDominatedCities));
end

function RealizeDominationStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		Controls.DominationViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.DominationViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
	else
		Controls.DominationViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.DominationViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
	end

	RealizeStackAndScrollbar(Controls.DominationViewStack, Controls.DominationViewScrollbar, true);
end

-- ===========================================================================
--	Called when Religion tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewReligion()
	ResetState(ViewReligion);
	Controls.ReligionView:SetHide(false);

	ChangeActiveHeader("VICTORY_RELIGIOUS", m_GenericHeaderIM, Controls.ReligionViewHeader);
	PopulateGenericHeader(RealizeReligionStackSize, RELIGION_TITLE, "", RELIGION_DETAILS, RELIGION_ICON);

	m_ReligionIM:ResetInstances();

	-- Pre-compute total number of Civs
	local numCivs:number = 0;
	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			numCivs = numCivs + 1;
		end
	end

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		-- Only show players with founded religions
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			local pReligion = pPlayer:GetReligion();
			if(pReligion ~= nil) then
				local religionType:number = pReligion:GetReligionTypeCreated();
				 if(religionType ~= -1) then
					PopulateReligionInstance(m_ReligionIM:GetInstance(), pPlayer, religionType, numCivs);
				end
			end
		end
	end

	RealizeReligionStackSize();
end

function PopulateReligionInstance(instance:table, pPlayer:table, iReligion:number, iTotalCivs:number)
	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateScoreInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	instance.CivName:SetText(civName);
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local religionData = GameInfo.Religions[iReligion];
	local religionColor:number = UI.GetColorValue(religionData.Color);
	
	instance.ReligionName:SetColor(religionColor);
	instance.ReligionName:SetText(Game.GetReligion():GetName(iReligion));
	instance.ReligionBG:SetSizeX(instance.ReligionName:GetSizeX() + PADDING_RELIGION_NAME_BG);
	instance.ReligionBG:SetColor(religionColor);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas("ICON_" .. religionData.ReligionType, SIZE_RELIGION_ICON_SMALL);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateReligionInstance: icon=\"".."ICON_" .. religionData.ReligionType.."\", iconSize="..tostring(SIZE_RELIGION_ICON_SMALL) );
	else
		instance.ReligionIcon:SetColor(religionColor);
		instance.ReligionIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
	end
	
	local convertedCivsIM:table = instance[DATA_FIELD_RELIGION_CONVERTED_CIVS_IM];
	if(convertedCivsIM == nil) then
		convertedCivsIM = InstanceManager:new("ConvertedReligionInstance", "CivIconBacking", instance.CivsConvertedStack);
		instance[DATA_FIELD_RELIGION_CONVERTED_CIVS_IM] = convertedCivsIM;
	end

	local iConvertedCivs:number = 0;
	convertedCivsIM:ResetInstances();
	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pOtherPlayer = Players[i];
		if(pOtherPlayer:IsAlive() == true and pOtherPlayer:IsMajor() == true) then
			local pReligion = pOtherPlayer:GetReligion();
			if(pReligion ~= nil) then
				local predominantReligion:number = pReligion:GetReligionInMajorityOfCities();
				if(predominantReligion == iReligion) then
					iConvertedCivs = iConvertedCivs + 1;

					local convertedInstance:table = convertedCivsIM:GetInstance();
					local iOtherPlayerID:number = pOtherPlayer:GetID();

					ColorCivIcon(convertedInstance, iOtherPlayerID);
					
					local civName:string, civIcon:string = GetCivNameAndIcon(iOtherPlayerID, true);
					textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, SIZE_RELIGION_CIV_ICON);
					if(textureSheet == nil or textureSheet == "") then
						UI.DataError("Could not find icon for converted religion in PopulateReligionInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_RELIGION_CIV_ICON));
					else
						convertedInstance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
						SetLeaderTooltip(convertedInstance, iOtherPlayerID);
					end
				end
			end
		end
	end

	if(iConvertedCivs == 0) then
		instance.ButtonBG:SetSizeY(SIZE_RELIGION_BG_HEIGHT + instance.CivsConvertedStack:GetSizeY());
	else
		instance.ButtonBG:SetSizeY(PADDING_RELIGION_BG_HEIGHT + instance.CivsConvertedStack:GetSizeY());
	end

	instance.CivsConverted:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_RELIGION_CONVERT_SUMMARY", iConvertedCivs .. "/" .. iTotalCivs, Game.GetReligion():GetName(iReligion)));
end

function RealizeReligionStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		Controls.ReligionViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.ReligionViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
	else
		Controls.ReligionViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.ReligionViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
	end

	RealizeStackAndScrollbar(Controls.ReligionViewStack, Controls.ReligionViewScrollbar, true);
end

-- ===========================================================================
--	Called when Generic (custom victory type) tab is selected (or when screen re-opens if selected)
-- ===========================================================================
function ViewGeneric(victoryType:string)
	ResetState(function() ViewGeneric(victoryType); end);
	Controls.GenericView:SetHide(false);

	ChangeActiveHeader("GENERIC", m_GenericHeaderIM, Controls.GenericViewHeader);

	local victoryInfo:table = GameInfo.Victories[victoryType];
	PopulateGenericHeader(RealizeGenericStackSize, victoryInfo.Name, nil, victoryInfo.Description, ICON_GENERIC);

	m_GenericIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateGenericInstance(m_GenericIM:GetInstance(), pPlayer, victoryType);
		end
	end

	RealizeGenericStackSize();
end

function PopulateGenericInstance(instance:table, pPlayer:table, victoryType:string)
	local playerID:number = pPlayer:GetID();
	local civName, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateViewAncientRivalsInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	instance.CivName:SetText(civName);
	instance.LocalPlayer:SetHide(pPlayer ~= g_LocalPlayer);

	local requirementSetID:number = Game.GetVictoryRequirements(playerID, victoryType);
	local innerRequirements:table = GameEffects.GetRequirementSetInnerRequirements(requirementSetID);

	local detailsText:string = "";
	for _, requirementID in ipairs(innerRequirements) do

		if detailsText ~= "" then
			detailsText = detailsText .. "[NEWLINE]";
		end

		local requirementKey:string = GameEffects.GetRequirementTextKey(requirementID, REQUIREMENT_CONTEXT);
		local requirementText:string = GameEffects.GetRequirementText(requirementID, requirementKey);

		if requirementText ~= nil then
			detailsText = detailsText .. requirementText;
		else
			local requirementState:string = GameEffects.GetRequirementState(requirementID);
			local requirementDetails:table = GameEffects.GetRequirementDefinition(requirementID);
			if requirementState == "Met" or requirementState == "AlwaysMet" then
				detailsText = detailsText .. "[ICON_CheckmarkBlue] ";
			else
				detailsText = detailsText .. "[ICON_Bolt]";
			end
			detailsText = detailsText .. requirementDetails.ID;
		end
	end
	instance.Details:SetText(detailsText);

	local itemSize:number = instance.Details:GetSizeY() + PADDING_GENERIC_ITEM_BG;
	if itemSize < SIZE_GENERIC_ITEM_MIN_Y then
		itemSize = SIZE_GENERIC_ITEM_MIN_Y;
	end
	
	instance.ButtonBG:SetSizeY(itemSize);
end

function RealizeGenericStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		Controls.GenericViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.GenericViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
	else
		Controls.GenericViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.GenericViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
	end

	RealizeStackAndScrollbar(Controls.GenericViewStack, Controls.GenericViewScrollbar, true);
end


-- // -------------------------------------------------------------------------------------------
-- // CiVI Reformation Victory functions
-- // -------------------------------------------------------------------------------------------

-- // Generic helper functions
-- // -------------------------------------------------------------------------------------------
function GetCustomVictoryProgressForPlayer(victoryType:string, player:table, index:number)

	local progress:number = 0;

	if (victoryType == "VICTORY_ECONOMIC") then
		progress = GetEconomicVictoryProgressRatio(player);
	elseif (victoryType == "VICTORY_METROPOLIS") then
		progress = CalculateMetropolisLargestCityProgress(player);
	elseif (victoryType == "VICTORY_TERRITORIAL") then
		progress = GetTerritorialVictoryProgressRatio(player);
	elseif (victoryType == "VICTORY_DIPLOMATIC") then
		progress = GetDiplomaticVictoryProgressRatio(player);
	else
		progress = Game.GetVictoryProgressForPlayer(victoryType, index); -- Game Core Call
	end
	
	if (progress > 1) then
		progress = 1;
	end
	
	return progress;
end

function SetVictoryBannerInfo(instance:table, icon:string, color:number)

	instance.VictoryBanner:SetColor(color);
	instance.VictoryLabelGradient:SetColor(color);
	
	if icon ~= nil then
		instance.VictoryIcon:SetTexture(icon);
		instance.VictoryIcon:SetHide(false);
	else
		instance.VictoryIcon:SetHide(true);
	end

end

function GetRequirementIconStateText(requirementState:boolean)

	local text:string = "[ICON_Bolt] ";
	
	if (requirementState == true) then
		text = "[ICON_CheckmarkBlue] ";
	end
	
	return text;
end	

function HasPlayerCompletedProject(player:table, project:string)

	local projectComplete:boolean = false;
	
	local playerStats:table = player:GetStats();
	local playerCities:table = player:GetCities();
	local projectInfo:table = GameInfo.Projects[project];
	
	for _, city in playerCities:Members() do

		local buildQueue:table = city:GetBuildQueue();
	
		if (buildQueue ~= nil) then
			local cost:number = buildQueue:GetProjectCost(projectInfo.Index);
			local progress:number = 0;
			
			if (playerStats:GetNumProjectsAdvanced(projectInfo.Index) == 0) then
				progress = buildQueue:GetProjectProgress(projectInfo.Index);
			else
				progress = cost;
			end
	
			if (progress >= cost) then
				projectComplete = true;
				break;
			end
		end
	
		if (projectComplete == true) then
			break;
		end
	end

	return projectComplete;
end

function GetMajorCivilizationsAlive(pPlayer:table)

	local numCivs:number = 0;
	
	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local player = Players[i];
		
		if (player:GetID() ~= pPlayer:GetID() and player:IsAlive() == true and player:IsMajor() == true) then
			numCivs = numCivs + 1;
		end
	end
	
	return numCivs;
end

function GetResearchProgressForTech(player:table, technologyKey:string)

	local progress:number = 0;
	
	if (player ~= nil) then
		local technology:table = GameInfo.Technologies[technologyKey];

		if (technology ~= nil) then
			local playerTechs = player:GetTechs();
			local currentTechID:number = playerTechs:GetResearchingTech();
			
			if (currentTechID == technology.Index) then
				local techProgress:number = playerTechs:GetResearchProgress(currentTechID);
				local techCost:number = playerTechs:GetResearchCost(currentTechID);
							
				if (techCost > 0) then
					progress = techProgress / techCost;
				end
			end
		end
	end
	
	return progress;
end

-- // Economic Victory functions
-- // -------------------------------------------------------------------------------------------
function ViewEconomic()
	ResetState(ViewEconomic);
	Controls.EconomicView:SetHide(false);

	ChangeActiveHeader("VICTORY_ECONOMIC", m_EconomicHeaderIM, Controls.EconomicViewHeader);
	PopulateGenericHeader(RealizeEconomicStackSize, ECONOMIC_TITLE, "", ECONOMIC_DETAILS, ECONOMIC_ICON);
	
	local text:string = "";
	
	text = text .. "[ICON_Bolt] " .. ECONOMIC_REQUIREMENT_GPT .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. ECONOMIC_REQUIREMENT_TRADE_ROUTES .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. ECONOMIC_REQUIREMENT_LUXURY_RESOURCES .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. ECONOMIC_REQUIREMENT_NATIONAL_BANK; -- .. "[NEWLINE][NEWLINE]";
	
	--text = text .. ECONOMIC_DETAILS_2;
	
	m_ActiveHeader.AdvisorTextCentered:SetText(text);
	m_ActiveHeader.HeaderIcon:SetTexture("Victory_Economic.dds");
	
	m_EconomicIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateEconomicInstance(m_EconomicIM:GetInstance(), pPlayer);
		end
	end

	RealizeEconomicStackSize();
end

function PopulateEconomicInstance(instance:table, pPlayer:table)
	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateEconomicInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	TruncateStringWithTooltip(instance.CivName, 180, civName); 
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local bHasMet = m_LocalPlayer:GetDiplomacy():HasMet(playerID);
	
	if (bHasMet == true or playerID == Game.GetLocalPlayer()) then
	
		local numCivs:number = GetMajorCivilizationsAlive(pPlayer);
	
		local hasGoldPerTurn:boolean = GetEconomicVictoryHasHighestGoldPerTurn(pPlayer);
		local hasTradeWithAllMajors:boolean = GetEconomicVictoryHasTradeWithAllMajors(pPlayer);
		local hasLuxuryResources:boolean = GetEconomicVictoryHasAllLuxuryResources(pPlayer);
		local hasNationalBank:boolean = GetEconomicVictoryHasNationalBank(pPlayer);
		
		-- // calculate and display gold per turn and current required minimum (total of all other civs)
		local playerGPT:number = GetEconomicVictoryGoldPerTurn(pPlayer);
		local playerGPTText:string = "[ICON_Gold][COLOR:ResGoldLabelCS]" .. Round(playerGPT, 1) .. "[ENDCOLOR]";
		local totalGPT:number = GetEconomicVictoryOpponentsGoldPerTurnTotal(pPlayer);
		local totalGPTText:string = "[ICON_Gold][COLOR:ResGoldLabelCS]" .. Round(totalGPT, 1) .. "[ENDCOLOR]";
		
		instance["GoldPerTurn"]:SetText(GetRequirementIconStateText(hasGoldPerTurn) .. "Gold Per Turn " .. playerGPTText .. " of " .. totalGPTText);
		
		-- // display the civilizations which we currently have a trade route to
		PopulateEconomicTradeInstance(instance, pPlayer, numCivs, hasTradeWithAllMajors);
		instance["TradePartnersStack"]:SetHide(false);
		
		PopulateEconomicLuxuryResourcesInstance(instance, pPlayer, m_MapLuxuryResourcesCount, hasLuxuryResources);
		instance["LuxuryResources"]:SetToolTipString(GetEconomicVictoryRequiredLuxuryResourcesTooltip(pPlayer, hasLuxuryResources));
		instance["LuxuryResourcesStack"]:SetHide(false);
		
		SetEconomicInstanceSize(instance);
	else	
		instance["GoldPerTurn"]:SetText("");
		instance["TradePartners"]:SetText("");
		instance["LuxuryResources"]:SetText("");
		
		instance.ButtonBG:SetSizeY(ECONOMIC_PANEL_DEFAULT_HEIGHT);
	end
end

function SetEconomicInstanceSize(instance:table)

	local tradePartnersStackHeight:number = math.max(instance["TradePartnersStack"]:GetSizeY(), ECONOMIC_PANEL_ICON_ROW_HEIGHT);
	local tradePartnersBufferHeight:number = 0;
	
	if (tradePartnersStackHeight > ECONOMIC_PANEL_ICON_ROW_HEIGHT) then
		-- we have more than one line of icons so increase the section size with the buffer height
		tradePartnersBufferHeight = ECONOMIC_PANEL_ICON_ROW_BUFFER;
	end
	
	local tradePartnersSectionHeight:number = tradePartnersStackHeight + tradePartnersBufferHeight + ECONOMIC_PANEL_SECTION_TEXT_HEIGHT;
	
	local luxuriesTextPositionY:number = 68 + tradePartnersStackHeight + tradePartnersBufferHeight; -- 68 = default size for civ icon and name
	local luxuriesStackHeight:number = math.max(instance["LuxuryResourcesStack"]:GetSizeY(), ECONOMIC_PANEL_ICON_ROW_HEIGHT);
	local luxuriesBufferHeight:number = 0;
	
	if (luxuriesStackHeight > ECONOMIC_PANEL_ICON_ROW_HEIGHT) then
		-- we have more than one line of icons so increase buffer
		luxuriesBufferHeight = ECONOMIC_PANEL_ICON_ROW_BUFFER;
	end
	
	local luxuriesSectionHeight:number = luxuriesStackHeight + luxuriesBufferHeight + ECONOMIC_PANEL_SECTION_TEXT_HEIGHT;
	
	instance["LuxuryResources"]:SetOffsetY(luxuriesTextPositionY); -- set the y position of the luxury resources text
	instance["LuxuryResourcesStack"]:SetOffsetY(luxuriesTextPositionY + ECONOMIC_PANEL_SECTION_TEXT_HEIGHT + 2); -- set the y position of the luxury resources icon stack
	instance.ButtonBG:SetSizeY(ECONOMIC_PANEL_DEFAULT_HEIGHT + tradePartnersSectionHeight + luxuriesSectionHeight); -- resize the panel to fit the sections
	
end

function PopulateEconomicTradeInstance(instance:table, pPlayer:table, numCivs:number, hasTradeRoutes:boolean)

	local tradePartnersIM:table = instance["TradePartnersIM"];
	
	if (tradePartnersIM == nil) then
		tradePartnersIM = InstanceManager:new("TradePartnersInstance", "CivIconBacking", instance.TradePartnersStack);
		instance["TradePartnersIM"] = tradePartnersIM;
	end
	
	if (tradePartnersIM ~= nil) then
		tradePartnersIM:ResetInstances();
	end
	
	local tradePartners:table = GetEconomicVictoryTradePartners(pPlayer);
	local numberOfTradePartners:number = 0;
	
	if (tradePartners ~= nil) then
		for tp, player in ipairs(tradePartners) do
	
			if (player:GetID() ~= pPlayer:GetID()) then
				numberOfTradePartners = numberOfTradePartners + 1;
			
				local tradePartnerInstance = tradePartnersIM:GetInstance();
				ColorCivIcon(tradePartnerInstance, player:GetID());
		
				local civName:string, civIcon:string = GetCivNameAndIcon(player:GetID(), true);
				textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, SIZE_RELIGION_CIV_ICON);
				
				if (textureSheet ~= nil) then
					tradePartnerInstance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
					SetLeaderTooltip(tradePartnerInstance, player:GetID());
				end
			end
			
		end
	end
	
	instance.TradePartners:SetText(GetRequirementIconStateText(hasTradeRoutes) .. Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_TRADE_ROUTES") .. " " .. numberOfTradePartners .. "/" .. numCivs);
end

function PopulateEconomicLuxuryResourcesInstance(instance:table, pPlayer:table, numLuxuries:number, hasLuxuryResources:boolean)

	local luxuryResourcesIM:table = instance["LuxuryResourcesIM"];
	
	if (luxuryResourcesIM == nil) then
		luxuryResourcesIM = InstanceManager:new("LuxuryResourcesInstance", "ResourceIcon", instance.LuxuryResourcesStack);
		instance["LuxuryResourcesIM"] = luxuryResourcesIM;
	end
	
	if (luxuryResourcesIM ~= nil) then
		luxuryResourcesIM:ResetInstances();
	end
	
	local luxuryResources:table = GetEconomicVictoryLuxuryResources(pPlayer);
	local numberOfLuxuryResources:number = 0;
	
	if (luxuryResources ~= nil) then
		for r, resource in ipairs(luxuryResources) do
			numberOfLuxuryResources = numberOfLuxuryResources + 1;
		
			local luxuryResourceInstance = luxuryResourcesIM:GetInstance();
			AddResourceIcon(luxuryResourceInstance, resource);
		end
	end
	
	instance.LuxuryResources:SetText(GetRequirementIconStateText(hasLuxuryResources) .. Locale.Lookup("LOC_WORLD_RANKINGS_ECONOMIC_REQUIREMENT_LUXURY_RESOURCES") .. " " .. numberOfLuxuryResources .. "/" .. numLuxuries);
end

function AddResourceIcon(instance:table, resource:table)			
				
	if (resource ~= nil) then
		local resourceName:string = Locale.Lookup(resource.Name);
		local resourceIcon:string = "ICON_" .. resource.ResourceType;
		
		local textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(resourceIcon, 38);
		
		if (textureSheet ~= nil) then
			instance.ResourceIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
			instance.ResourceIcon:SetToolTipString(resourceName);
		end
	end
	
end

function GetEconomicVictoryRequiredLuxuryResourcesTooltip(player:table, hasLuxuryResources:boolean)

	local tooltip:string = "";

	if (not hasLuxuryResources and player:GetID() == m_LocalPlayerID) then -- only show list of missing resources for player
		if (m_LocalPlayerRequiredLuxuryResources ~= nil) then
			tooltip = "Required Resources:[NEWLINE]";
			
			for i, resource in ipairs(m_LocalPlayerRequiredLuxuryResources) do
				local resourceIcon:string = "[ICON_" .. resource.ResourceType .. "]";
				
				tooltip = tooltip .. "[NEWLINE]" .. resourceIcon .. " " .. Locale.Lookup(resource.Name);
			end
		end		
	end

	return tooltip;
end

function RealizeEconomicStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		
		m_ActiveHeader.AdvisorTextCentered:SetOffsetY(OFFSET_SCIENCE_REQUIREMENTS_Y + m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_HEADER);
		headerHeight = headerHeight + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + (PADDING_HEADER * 2);
		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);
		
		Controls.EconomicViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.EconomicViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(false);
	else
		Controls.EconomicViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.EconomicViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.EconomicViewStack, Controls.EconomicViewScrollbar, true);
end

-- // Economic helper functions
-- // -------------------------------------------------------------------------------------------
function GetEconomicVictoryProgressRatio(player:table)

	local ratio:number = 0;

	ratio = GetEconomicVictoryGoldPerTurnRatio(player) * 0.8; -- accounts for 80% of icon progress

	if (GetEconomicVictoryHasTradeWithAllMajors(player)) then
		ratio = ratio + 0.1; -- accounts for 10% of icon progress. may change in future version
	end
	
	if (GetEconomicVictoryHasAllLuxuryResources(player)) then
		ratio = ratio + 0.1; -- accounts for 10% of icon progress. may change in future version
	end
	
	return ratio; 
end

function GetEconomicVictoryGoldPerTurn(player:table)

	local goldPerTurn:number = 0;

	if (player ~= nil) then
		goldPerTurn = player:GetTreasury():GetGoldYield();
		goldPerTurn = goldPerTurn - player:GetTreasury():GetTotalMaintenance();		
	end
	
	return goldPerTurn;
end

function GetEconomicVictoryGoldPerTurnTotal()

	local goldPerTurn:number = 0;

	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		goldPerTurn = goldPerTurn + GetEconomicVictoryGoldPerTurn(player);
	end
	
	return goldPerTurn;
end

function GetEconomicVictoryOpponentsGoldPerTurnTotal(currentPlayer:table)

	local goldPerTurn:number = 0;

	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		if (currentPlayer:GetID() ~= player:GetID()) then
			goldPerTurn = goldPerTurn + GetEconomicVictoryGoldPerTurn(player);
		end
	end
	
	return goldPerTurn;
end

function GetEconomicVictoryGoldPerTurnRatio(player:table)

	local ratio:number = 0;
	
	local playerGoldPerTurn:number = GetEconomicVictoryGoldPerTurn(player);
	
	if (playerGoldPerTurn > 0) then
		local goldPerTurnTotal:number = GetEconomicVictoryGoldPerTurnTotal();
	
		if (goldPerTurnTotal > 0) then
			if (playerGoldPerTurn >= goldPerTurnTotal) then
				ratio = 1;
			else
				ratio = playerGoldPerTurn / goldPerTurnTotal;
			end
		
			if (ratio > 1) then
				ratio = 1;
			end
		end
	end
	
	return ratio;
end

function GetEconomicVictoryHasNationalBank(player:table)

	local projectCompleted:boolean = false;

	if (player ~= nil) then
		projectCompleted = HasPlayerCompletedProject(player, "PROJECT_NATIONAL_BANK");
	end

	return projectCompleted;
end

function GetEconomicVictoryHasHighestGoldPerTurn(currentPlayer:table)

	local result:boolean = false;
	
	local playerGoldPerTurn:number = GetEconomicVictoryGoldPerTurn(currentPlayer);
	local opponentsGoldPerTurn:number = 0;

	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		local currentPlayerID:number = currentPlayer:GetID();
		local playerID:number = player:GetID();
	
		if (currentPlayerID ~= playerID) then
			opponentsGoldPerTurn = opponentsGoldPerTurn + GetEconomicVictoryGoldPerTurn(player);
		end
	end

	if (playerGoldPerTurn >= opponentsGoldPerTurn) then
		result = true;
	end
	
	return result;
end

function GetPlayerCivilizationName(playerID:number)

	local name:string = "";
	local config:table = PlayerConfigurations[playerID];
	
	if (config ~= nil) then
		name = config:GetCivilizationTypeName();
	end
	
	return name;
end

function GetEconomicVictoryTradePartners(currentPlayer:table)

	-- // determine the collection of trade routes for the current player
	local tradeRoutes = {};
	local playerCities:table = currentPlayer:GetCities();
		
	if (playerCities ~= nil) then
	
		for _, city in playerCities:Members() do
			local outgoingRoutes:table = city:GetTrade():GetOutgoingRoutes();
		
			for i, route in ipairs(outgoingRoutes) do
				table.insert(tradeRoutes, route);
			end
		end
		
	end
	
	-- // determine the collection of civilizations for the trade routes
	local tradePartners = {};
	
	for r, route in ipairs(tradeRoutes) do
		local destinationCityPlayerID:number = route.DestinationCityPlayer;
		
		if (Players[destinationCityPlayerID]:IsMajor()) then
			if (not IsPlayerInTradePartnersTable(destinationCityPlayerID, tradePartners)) then
				table.insert(tradePartners, Players[destinationCityPlayerID]);
			end
		end
	end

	return tradePartners;
end

function GetEconomicVictoryHasTradeWithAllMajors(currentPlayer:table)

	-- // we assume that we have a trade route with every major civilization
	local tradeWithAllMajors:boolean = true;
		
	-- // determine the collection of trade routes for the current player
	local tradeRoutes = {};
	local playerCities:table = currentPlayer:GetCities();
		
	if (playerCities ~= nil) then
	
		for _, city in playerCities:Members() do
			local outgoingRoutes:table = city:GetTrade():GetOutgoingRoutes();
		
			for i, route in ipairs(outgoingRoutes) do
				table.insert(tradeRoutes, route);
			end
		end
		
	end
	
	-- // determine the collection of civilizations for the trade routes
	local tradePartners = {};
	
	for r, route in ipairs(tradeRoutes) do
		local destinationCityPlayerID:number = route.DestinationCityPlayer;
		
		if (not IsPlayerInTradePartnersTable(destinationCityPlayerID, tradePartners)) then
			table.insert(tradePartners, Players[destinationCityPlayerID]);
		end
	end
	
	-- // determine if all major civilizations have a trade route
	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		local playerID:number = player:GetID();
		
		if (playerID ~= currentPlayer:GetID() and not IsPlayerInTradePartnersTable(playerID, tradePartners)) then
			tradeWithAllMajors = false;
			break;
		end
	end
	
	return tradeWithAllMajors;
end

function IsPlayerInTradePartnersTable(playerID:number, tradePartners:table)

	local result:boolean = false;

	if (tradePartners ~= nil) then
		for tp, player in ipairs(tradePartners) do
			if (player:GetID() == playerID) then
				result = true; -- player already exists in table
				break;
			end
		end
	end
	
	return result;
end

function LoadMapLuxuryResourceData()

	print("LoadMapLuxuryResourceData: ENTER");

	if (m_MapLuxuryResourcesCount == 0) then
		local mapX, mapY = Map.GetGridSize();
		
		-- loop through the columns on the map grid
		for x = 0, mapX - 1 do
			-- now loop through the rows for each column on the map grid
			for y = 0, mapY - 1 do
				local plot:table = Map.GetPlot(x, y); -- get the plot at the x and y position
				local resource:table = GetLuxuryResourceAtPlot(plot); -- get luxury resource if one exists at plot
				
				if (resource ~= nil) then				
					if (resource.Name ~= nil and not IsResourceInLuxuryResourcesTable(resource.Name)) then
						-- resource does not exist in table of map resources, add it and increase the counter
						table.insert(m_MapLuxuryResources, resource);
						m_MapLuxuryResourcesCount = m_MapLuxuryResourcesCount + 1;
					
						print("LoadMapLuxuryResourceData: resource ", resource.Name);					
						print("LoadMapLuxuryResourceData: new resource added to m_MapLuxuryResources collection");
						print("LoadMapLuxuryResourceData: new resources found = ", m_MapLuxuryResourcesCount);
					end
				end
			end
		end
	end

	print("LoadMapLuxuryResourceData: EXIT");
end

function GetLuxuryResourceAtPlot(plot:table)

	local luxuryResource:table = {};

	if (plot ~= nil and plot:GetResourceCount() > 0) then
		local resourceType:number = plot:GetResourceType();
		
		if (resourceType ~= -1) then
			local resource = GameInfo.Resources[resourceType];
			
			if (resource ~= nil and resource.ResourceClassType == "RESOURCECLASS_LUXURY") then
				luxuryResource = resource;
			end
		end
	end
	
	return luxuryResource;
end

function IsResourceInLuxuryResourcesTable(resourceName:string)

	local result:boolean = false;

	if (m_MapLuxuryResources ~= nil) then
		for r, resource in ipairs(m_MapLuxuryResources) do
			if (resource.Name == resourceName) then
				result = true; -- resource already exists in table
				break;
			end
		end
	end
	
	return result;
end

function GetEconomicVictoryLuxuryResources(player:table)

	local luxuryResources:table = {};

	for r, resource in ipairs(m_MapLuxuryResources) do
		if (player:GetResources():HasResource(resource.Index)) then
			table.insert(luxuryResources, resource);
		end
	end
	
	return luxuryResources;
end

function GetEconomicVictoryHasAllLuxuryResources(player:table)

	local result:boolean = false;
	
	local totalLuxuryResources:number = m_MapLuxuryResourcesCount;
	local playerLuxuryResources:number = 0;
	local isLocalPlayer:boolean = (player:GetID() == m_LocalPlayerID);
		
	if (m_MapLuxuryResources ~= nil) then
		if (isLocalPlayer) then
			m_LocalPlayerRequiredLuxuryResources = {};
		end
	
		for r, resource in ipairs(m_MapLuxuryResources) do
		
			if (player:GetResources():HasResource(resource.Index)) then
				playerLuxuryResources = playerLuxuryResources + 1;
			else
				if (isLocalPlayer) then
					table.insert(m_LocalPlayerRequiredLuxuryResources, resource);
				end
			end
			
		end
		
		if (isLocalPlayer and m_LocalPlayerRequiredLuxuryResources ~= nil) then
			table.sort(m_LocalPlayerRequiredLuxuryResources, function(a, b) return Locale.Compare(a.Name, b.Name) == -1; end);
		end
	end

	-- // if the player has access to all luxury resources then return true
	if (playerLuxuryResources >= totalLuxuryResources) then
		result = true;
	end
	
	return result;
end

function GetEconomicVictoryCompleted(player:table)

	local result:boolean = false;

	if (GetEconomicVictoryHasHighestGoldPerTurn(player)) then
		if (GetEconomicVictoryHasAllLuxuryResources(player)) then
			if (GetEconomicVictoryHasTradeWithAllMajors(player)) then
				result = GetEconomicVictoryHasNationalBank(player);
			end
		end
	end
	
	return result;
end

-- // Metropolis Victory functions
-- // -------------------------------------------------------------------------------------------
function ViewMetropolis()
	ResetState(ViewMetropolis);
	Controls.MetropolisView:SetHide(false);

	ChangeActiveHeader("VICTORY_METROPOLIS", m_MetropolisHeaderIM, Controls.MetropolisViewHeader);
	PopulateGenericHeader(RealizeMetropolisStackSize, METROPOLIS_TITLE, "", METROPOLIS_DETAILS, METROPOLIS_ICON);

	local text:string = "";
	
	text = text .. "[ICON_Bolt] " .. METROPOLIS_REQUIREMENT_HIGH_POP_CITY .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. METROPOLIS_REQUIREMENT_HAPPY; -- .. "[NEWLINE]";
	--text = text .. "[ICON_Bolt] " .. METROPOLIS_REQUIREMENT_MONUMENT .. "[NEWLINE][NEWLINE]";
	
	--text = text .. METROPOLIS_DETAILS_2;
	
	m_ActiveHeader.AdvisorTextCentered:SetText(text);
	m_ActiveHeader.HeaderIcon:SetTexture("Victory_Metropolis.dds");
	
	m_MetropolisIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateMetropolisInstance(m_MetropolisIM:GetInstance(), pPlayer);
		end
	end

	RealizeMetropolisStackSize();	
end

function PopulateMetropolisInstance(instance:table, pPlayer:table)

	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateMetropolisInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end
	
	local amenitiesTextureOffsetX:number, amenitiesTextureOffsetY:number, amenitiesTextureSheet:string = IconManager:FindIconAtlas("ICON_AMENITIES", 22);
	
	if (amenitiesTextureSheet ~= nil) then
		instance.AmenitiesIcon:SetTexture(amenitiesTextureOffsetX, amenitiesTextureOffsetY, amenitiesTextureSheet);
	end

	ColorCivIcon(instance, playerID);
	TruncateStringWithTooltip(instance.CivName, 180, civName); 
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local bHasMet = m_LocalPlayer:GetDiplomacy():HasMet(playerID);
	
	if (bHasMet == true or playerID == Game.GetLocalPlayer()) then
	
		local population:number = CalculateMetropolisLargestCity(pPlayer);
		local playerCity:table = GetMetropolisVictoryHighestPopulationCity(pPlayer);
		local percentage:number = 0;
		local cityHappiness:number = 0;
		
		if (playerCity == nil) then		
			instance.PopulationBar:SetPercent(percentage);
		else
			cityHappiness = GetCityHappiness(playerCity);
			percentage = population / METROPOLIS_POPULATION_TARGET;
			
			if (percentage > 1) then
				percentage = 1;
			end
			
			instance.PopulationBar:SetPercent(percentage);
		end
		
		SetAmenitiesColour(instance, cityHappiness);
		
		local targetPopulationReached:boolean = population >= METROPOLIS_POPULATION_TARGET;
		local targetHappinessReached:boolean = cityHappiness >= METROPOLIS_MIN_HAPPINESS;
		--local playerCompletedProject:boolean = HasPlayerCompletedProject(pPlayer, "PROJECT_METROPOLIS_MONUMENT");
		
		instance.PopulationBorder:SetToolTipString(GetMetropolisVictoryPopulationTooltip(playerCity, targetPopulationReached));
		instance.AmenitiesIcon:SetToolTipString(GetMetropolisVictoryHappinessTooltip(playerCity, targetHappinessReached));
	else
		--SetAmenitiesColour(instance, 0);
		instance["PopulationBar"]:SetPercent(0);
		instance["PopulationBorder"]:SetToolTipString("");
		instance["AmenitiesIcon"]:SetToolTipString("");
		
		instance.AmenitiesIcon:SetColor(0x55555555);
	end
end

function GetMetropolisVictoryPopulationTooltip(city:table, targetPopulationReached:boolean)

	local tooltip:string = "";

	if (city:IsCapital()) then
		tooltip = tooltip .. "[ICON_Capital]";
	end
	
	tooltip = tooltip .. Locale.ToUpper(Locale.Lookup(city:GetName())) .. "[NEWLINE]";
	
	tooltip = tooltip .. GetRequirementIconStateText(targetPopulationReached);
	tooltip = tooltip .. Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_TOOLTIP_POPULATION") .. " " .. city:GetPopulation();
	
	return tooltip;
end

function GetMetropolisVictoryHappinessTooltip(city:table, targetHappinessReached:boolean)

	local tooltip:string = "";
	local happiness:number = GetCityHappiness(city);

	tooltip = tooltip .. GetRequirementIconStateText(targetHappinessReached);
	tooltip = tooltip .. Locale.Lookup("LOC_WORLD_RANKINGS_METROPOLIS_TOOLTIP_HAPPINESS");
	tooltip = tooltip .. " " .. Locale.Lookup(GameInfo.Happinesses[happiness].Name);	
	
	return tooltip;
end

function RealizeMetropolisStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		
		m_ActiveHeader.AdvisorTextCentered:SetOffsetY(OFFSET_SCIENCE_REQUIREMENTS_Y + m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_HEADER);
		headerHeight = headerHeight + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + (PADDING_HEADER * 2);
		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);
		
		Controls.MetropolisViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.MetropolisViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(false);
	else
		Controls.MetropolisViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.MetropolisViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.MetropolisViewStack, Controls.MetropolisViewScrollbar, true);
end

-- // Metropolis helper functions
-- // -------------------------------------------------------------------------------------------
function SetAmenitiesColour(instance:table, cityHappiness:number)

	local colourSetName:string = "StatNormalCS";
	local happinessInfo = GameInfo.Happinesses[cityHappiness];
	
	if (happinessInfo ~= nil) then
		if (happinessInfo.GrowthModifier < 0) then
			colourSetName = "StatBadCS";
		elseif (happinessInfo.GrowthModifier > 0) then
			colourSetName = "StatGoodCS";
		end
	end

	instance.AmenitiesIcon:SetColorByName(colourSetName);
	
end

function GetCityHappiness(city:table)

	local happiness:number = 0;
	
	if (city ~= nil) then
		local cityGrowth:table = city:GetGrowth();
	
		if (cityGrowth ~= nil) then
			happiness = cityGrowth:GetHappiness();
		end
	end

	return happiness;
end

function GetMetropolisVictoryHighestPopulationCity(player:table)

	local largestCity:table = {};	
	
	if (player ~= nil) then
	
		local largestCityPopulation:number = 0;
		local playerCities:table = player:GetCities();
		
		if (playerCities ~= nil) then
			for _, city in playerCities:Members() do
				local cityPopulation = city:GetPopulation();
	
				if (cityPopulation > largestCityPopulation) then
					largestCityPopulation = cityPopulation;
					largestCity = city;
				end
			end
		end
	
	end

	return largestCity;
end

function CalculateMetropolisLargestCity(player:table)

	local playerCities:table = player:GetCities();
	local largestCitySize:number = 0;
	local citySize:number = 0;
	
	for _, city in playerCities:Members() do
		citySize = city:GetPopulation();
	
		if (citySize > largestCitySize) then
			largestCitySize = citySize;
		end	
	end
	
	return largestCitySize;
end

function CalculateMetropolisLargestCityProgress(player:table)

	local progress:number = 0;
	local population:number = CalculateMetropolisLargestCity(player);
	
	if (population > 0) then
		progress = population / METROPOLIS_POPULATION_TARGET;
	end
	
	return progress;
end

function GetMetropolisVictoryCompleted(player:table, targetHappiness:number)

	local playerCity:table = GetMetropolisVictoryHighestPopulationCity(player);
	local metropolisVictoryCompleted:boolean = false;
	
	if (playerCity ~= nil) then
		local population:number = CalculateMetropolisLargestCity(player);
	
		if (population >= METROPOLIS_POPULATION_TARGET) then
			local happiness:number = GetCityHappiness(playerCity);
		
			if (happiness >= targetHappiness) then			
				-- // we have achieved a Metropolis victory as we have a city with the required
				-- // population and the citizens of the city are Happy or Ecstatic. 
				metropolisVictoryCompleted = true;
			end
		end
		
	end

	return metropolisVictoryCompleted;
end

-- // Territorial Victory functions
-- // -------------------------------------------------------------------------------------------
function ViewTerritorial()
	ResetState(ViewTerritorial);
	Controls.TerritorialView:SetHide(false);

	ChangeActiveHeader("VICTORY_TERRITORIAL", m_TerritorialHeaderIM, Controls.TerritorialViewHeader);
	PopulateGenericHeader(RealizeTerritorialStackSize, TERRITORIAL_TITLE, "", TERRITORIAL_DETAILS, TERRITORIAL_ICON);

	local text:string = "";
	
	text = text .. "[ICON_Bolt] " .. TERRITORIAL_REQUIREMENT_MOST_TERRITORY .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. TERRITORIAL_REQUIREMENT_LARGEST_MILITARY .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. TERRITORIAL_REQUIREMENT_TELECOMMUNICATIONS; -- .. "[NEWLINE]";
	--text = text .. "[ICON_Bolt] " .. TERRITORIAL_REQUIREMENT_MONUMENT .. "[NEWLINE][NEWLINE]";
	
	--text = text .. TERRITORIAL_DETAILS_2;
	
	m_ActiveHeader.AdvisorTextCentered:SetText(text);
	m_ActiveHeader.HeaderIcon:SetTexture("Victory_Territorial.dds");
	
	m_TerritorialIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateTerritorialInstance(m_TerritorialIM:GetInstance(), pPlayer);
		end
	end

	RealizeTerritorialStackSize();
end

function PopulateTerritorialInstance(instance:table, pPlayer:table)

	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateTerritorialInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	TruncateStringWithTooltip(instance.CivName, 180, civName);
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local bHasMet = m_LocalPlayer:GetDiplomacy():HasMet(playerID);
	
	if (bHasMet == true or playerID == Game.GetLocalPlayer()) then
		
		local territorySize:number = GetTerritorialVictoryTerritorySize(pPlayer);
		local territoryRatio:number = GetCustomVictoryProgressForPlayer("VICTORY_TERRITORIAL", pPlayer, 0); --GetTerritorialVictoryTerritoryRatio(territorySize);
		local telecommunicationsProgress = GetResearchProgressForTech(pPlayer, "TECH_TELECOMMUNICATIONS");
		
		local playerHasMostTerritory:boolean = GetTerritorialVictoryHasMostTerritory(pPlayer, territorySize);
		local playerHasLargestMilitary:boolean = GetTerritorialVictoryHasLargestMilitary(pPlayer);
		local playerHasTelecommunications:boolean = GetTerritorialVictoryHasTechnology(pPlayer, "TECH_TELECOMMUNICATIONS");
		--local playerHasCompletedProject:boolean = HasPlayerCompletedProject(pPlayer, "PROJECT_TERRITORIAL_MONUMENT");
		
		if (playerHasTelecommunications) then
			telecommunicationsProgress = 1;
		end
		
		instance.TerritoryBar:SetPercent(territoryRatio);
		instance.ResearchBar:SetPercent(telecommunicationsProgress);
		
		instance.MilitaryIcon:SetHide(false);
				
		if (playerHasLargestMilitary) then
			instance.MilitaryIcon:SetColor(0xdddddddd);
		else
			instance.MilitaryIcon:SetColor(0x55555555);
		end
		
		instance.TerritoryBorder:SetToolTipString(GetTerritorialVictoryTerritorySizeTooltip(playerHasMostTerritory, territoryRatio));
		instance.ResearchBorder:SetToolTipString(GetTerritorialVictoryResearchTooltip(playerHasTelecommunications));
		instance.MilitaryIcon:SetToolTipString(GetTerritorialVictoryMilitaryTooltip(playerHasLargestMilitary));
	else				
		instance.TerritoryBar:SetPercent(0);
		instance.ResearchBar:SetPercent(0);
		--instance.MilitaryIcon:SetHide(true);
		instance.MilitaryIcon:SetColor(0x55555555);
		
		instance.TerritoryBorder:SetToolTipString("");
		instance.ResearchBorder:SetToolTipString("");
		instance.MilitaryIcon:SetToolTipString("");
	end
end

function GetTerritorialVictoryTerritorySizeTooltip(playerHasMostTerritory:boolean, territoryRatio:number)

	local tooltip:string = "";

	tooltip = tooltip .. GetRequirementIconStateText(playerHasMostTerritory);
	tooltip = tooltip .. Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_TOOLTIP_MOST_TERRITORY");
	tooltip = tooltip .. " (" .. Round(territoryRatio * 100, 1) .. "%)";
	
	return tooltip;
end

function GetTerritorialVictoryMilitaryTooltip(playerHasLargestMilitary:boolean)

	local tooltip:string = "";

	tooltip = tooltip .. GetRequirementIconStateText(playerHasLargestMilitary);
	tooltip = tooltip .. Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_TOOLTIP_LARGEST_MILITARY");	
	
	return tooltip;
end

function GetTerritorialVictoryResearchTooltip(playerHasTelecommunications:boolean)

	local tooltip:string = "";

	tooltip = tooltip .. GetRequirementIconStateText(playerHasTelecommunications);
	tooltip = tooltip .. Locale.Lookup("LOC_WORLD_RANKINGS_TERRITORIAL_TOOLTIP_TELECOMMUNICATIONS");	
	
	return tooltip;
end

function RealizeTerritorialStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		
		m_ActiveHeader.AdvisorTextCentered:SetOffsetY(OFFSET_SCIENCE_REQUIREMENTS_Y + m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_HEADER);
		headerHeight = headerHeight + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + (PADDING_HEADER * 2);
		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);
		
		Controls.TerritorialViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.TerritorialViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(false);
	else
		Controls.TerritorialViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.TerritorialViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.TerritorialViewStack, Controls.TerritorialViewScrollbar, true);
end

-- // Territorial helper functions
-- // -------------------------------------------------------------------------------------------
function GetTerritorialVictoryHasMinimumTerritorySize(player:table, territorySize:number)

	local result:boolean = false;
	local percentage:number = GetTerritorialVictoryTerritorySizeAsPercentage(player, territorySize);
	
	if (percentage >= TERRITORIAL_MIN_EMPIRE_SIZE) then
		result = true;
	end
	
	return result;
end

function GetTerritorialVictoryProgressRatio(player:table)

	local ratio:number = 0;

	local territorySize:number = GetTerritorialVictoryTerritorySize(player);
	local territoryRatio:number = GetTerritorialVictoryTerritoryRatio(territorySize);
	
	return territoryRatio;
end

function GetTerritorialVictoryTerritorySize(player:table)

	local size:number = 0;

	if (player ~= nil) then
		local playerCities:table = player:GetCities();
		
		for _, city in playerCities:Members() do
			local plots:table = Map.GetCityPlots():GetPurchasedPlots(city);
		
			for _, plotID in ipairs(plots) do
				size = size + 1;
			end
		end
	end
	
	return size;
end

function GetTerritorialVictoryTerritoryRatio(territorySize:number)

	local ratio:number = 0;
	local totalTerritoryOwned:number = GetTerritorialVictoryTotalTerritory();
	
	if (totalTerritoryOwned > 0) then
		ratio = territorySize / totalTerritoryOwned;
		
		if (ratio > 1) then
			ratio = 1;
		end
	end
	
	return ratio;
end

function GetTerritorialVictoryTerritorySizeAsPercentage(player:table)

	local territoryOwned:number = GetTerritorialVictoryTerritorySize(player);
	local territoryRatio:number = GetTerritorialVictoryTerritoryRatio(territoryOwned);
	
	return territoryRatio * 100;
end

function GetTerritorialVictoryTotalTerritory()

	local size:number = 0;

	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		size = size + GetTerritorialVictoryTerritorySize(player);
	end
	
	return size;
end

function GetTerritorialVictoryHasMostTerritory(currentPlayer:table, territorySize:number)

	local result:boolean = true;
	
	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		local currentPlayerID:number = currentPlayer:GetID();
		local playerID:number = player:GetID();
	
		if (currentPlayerID ~= playerID) then
			local size:number = GetTerritorialVictoryTerritorySize(player);
			
			if (size > territorySize) then
				result = false;
				break;
			end
		end
	end
	
	if (result == true) then
		if (not GetTerritorialVictoryHasMinimumTerritorySize(currentPlayer, territorySize)) then
			result = false; -- if we are leading but do not have the minimum required, set return value to false
		end
	end

	return result;
end

function GetTerritorialVictoryHasLargestMilitary(currentPlayer:table)

	local result:boolean = true;
	local currentPlayerStrength:number = currentPlayer:GetStats():GetMilitaryStrength();
	
	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		local currentPlayerID:number = currentPlayer:GetID();
		local playerID:number = player:GetID();
	
		if (currentPlayerID ~= playerID) then
			local strength:number = player:GetStats():GetMilitaryStrength();
			
			if (strength > currentPlayerStrength) then
				result = false;
				break;
			end
		end
	end

	return result;
end

function GetTerritorialVictoryHasTechnology(player:table, technologyKey:string)

	local result:boolean = false;

	if (player ~= nil) then
		local technology:table = GameInfo.Technologies[technologyKey];
	
		if (technology ~= nil) then
			result = player:GetTechs():HasTech(technology.Index);
		end
	end
	
	return result;
end

function GetTerritorialVictoryCompleted(player:table)

	local result:boolean = false;

	if (GetTerritorialVictoryHasTechnology(player, "TECH_TELECOMMUNICATIONS")) then
		if (GetTerritorialVictoryHasLargestMilitary(player)) then
			local territorySize:number = GetTerritorialVictoryTerritorySize(player);
			result = GetTerritorialVictoryHasMostTerritory(player, territorySize);
		end
	end
	
	return result;
end

-- // Diplomatic Victory functions
-- // -------------------------------------------------------------------------------------------
function ViewDiplomatic()
	ResetState(ViewDiplomatic);
	Controls.DiplomaticView:SetHide(false);

	ChangeActiveHeader("VICTORY_DIPLOMATIC", m_DiplomaticHeaderIM, Controls.DiplomaticViewHeader);
	PopulateGenericHeader(RealizeDiplomaticStackSize, DIPLOMATIC_TITLE, "", DIPLOMATIC_DETAILS, DIPLOMATIC_ICON);
	
	local text:string = "";

	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_MAJOR_PEACE .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_MAJOR_EMBASSY .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_MAJOR_ALLY .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_MINOR_PEACE .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_MINOR_SUZERAIN .. "[NEWLINE]";
	text = text .. "[ICON_Bolt] " .. DIPLOMATIC_REQUIREMENT_WORLD_COUNCIL;
	
	m_ActiveHeader.AdvisorTextCentered:SetText(text);
	m_ActiveHeader.HeaderIcon:SetTexture("Victory_Diplomatic.dds");
	
	m_DiplomaticIM:ResetInstances();

	for i = 0, PlayerManager.GetWasEverAliveCount() - 1 do
		local pPlayer = Players[i];
		if (pPlayer:IsAlive() == true and pPlayer:IsMajor() == true) then
			PopulateDiplomaticInstance(m_DiplomaticIM:GetInstance(), pPlayer);
		end
	end

	RealizeDiplomaticStackSize();
end

function PopulateDiplomaticInstance(instance:table, pPlayer:table)
	local playerID:number = pPlayer:GetID();
	local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);

	local textureOffsetX:number, textureOffsetY:number, textureSheet:string = IconManager:FindIconAtlas(civIcon, SIZE_CIV_ICON);
	
	if(textureSheet == nil or textureSheet == "") then
		UI.DataError("Could not find icon in PopulateDiplomaticInstance: icon=\""..civIcon.."\", iconSize="..tostring(SIZE_CIV_ICON));
	else
		instance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
		SetLeaderTooltip(instance, playerID);
	end

	ColorCivIcon(instance, playerID);
	TruncateStringWithTooltip(instance.CivName, 180, civName); 
	instance.LocalPlayer:SetHide(pPlayer ~= m_LocalPlayer);

	local bHasMet = m_LocalPlayer:GetDiplomacy():HasMet(playerID);
	
	if (bHasMet == true or playerID == Game.GetLocalPlayer()) then
		-- // display the major civilization icons
		PopulateDiplomaticMajorCivsInstance(instance, pPlayer);
		instance["MajorCivsStack"]:SetHide(false);

		-- // display the minor civilization icons
		PopulateDiplomaticMinorCivsInstance(instance, pPlayer);
		instance["MinorCivsStack"]:SetHide(false);
		
		PopulateDiplomaticWorldCouncilLabel(instance, pPlayer);
		
		SetDiplomaticInstanceSize(instance);
	else
		instance["WorldCouncilLabel"]:SetText("");
		instance["MajorCivsLabel"]:SetText("");
		instance["MinorCivsLabel"]:SetText("");
		
		instance.ButtonBG:SetSizeY(DIPLOMATIC_PANEL_DEFAULT_HEIGHT);
	end
end

function SetDiplomaticInstanceSize(instance:table)

	-- // define sizes of elements in the major civs section of the panel
	-- // ----------------------------------------------------------------
	local majorCivsStackHeight:number = math.max(instance["MajorCivsStack"]:GetSizeY(), DIPLOMATIC_PANEL_ICON_ROW_HEIGHT);
	local majorCivsBufferHeight:number = 0;

	if (majorCivsStackHeight > DIPLOMATIC_PANEL_ICON_ROW_HEIGHT) then
		-- we have more than one line of icons so increase the section size with the buffer height
		majorCivsBufferHeight = DIPLOMATIC_PANEL_ICON_ROW_BUFFER;
	end
	
	local majorCivsSectionHeight:number = majorCivsStackHeight + majorCivsBufferHeight + DIPLOMATIC_PANEL_SECTION_TEXT_HEIGHT;
	
	-- // define sizes of elements in the minor civs section of the panel
	-- // ----------------------------------------------------------------
	local minorCivsTextPositionY:number = 68 + majorCivsStackHeight + majorCivsBufferHeight; -- 68 = default size for civ icon and name
	local minorCivsStackHeight:number = math.max(instance["MinorCivsStack"]:GetSizeY(), DIPLOMATIC_PANEL_ICON_ROW_HEIGHT);
	local minorCivsBufferHeight:number = 0;

	if (minorCivsStackHeight > DIPLOMATIC_PANEL_ICON_ROW_HEIGHT) then
		-- we have more than one line of icons so increase buffer
		minorCivsBufferHeight = DIPLOMATIC_PANEL_ICON_ROW_BUFFER;
	end
	
	local minorCivsSectionHeight:number = minorCivsStackHeight + minorCivsBufferHeight + DIPLOMATIC_PANEL_SECTION_TEXT_HEIGHT;
	
	instance["MinorCivsLabel"]:SetOffsetY(minorCivsTextPositionY); -- set the y position of the minor civilizations text
	instance["MinorCivsStack"]:SetOffsetY(minorCivsTextPositionY + DIPLOMATIC_PANEL_SECTION_TEXT_HEIGHT + 2); -- set the y position of the minor civs icon stack
	instance.ButtonBG:SetSizeY(DIPLOMATIC_PANEL_DEFAULT_HEIGHT + majorCivsSectionHeight + minorCivsSectionHeight); -- resize the panel to fit the sections
	
end

function PopulateDiplomaticMajorCivsInstance(instance:table, currentPlayer:table)
	
	local majorCivsIM:table = instance["MajorCivsIM"];
	
	if (majorCivsIM == nil) then
		majorCivsIM = InstanceManager:new("MajorCivsInstance", "CivIconBacking", instance.MajorCivsStack);
		instance["MajorCivsIM"] = majorCivsIM;
	end
	
	if (majorCivsIM ~= nil) then
		majorCivsIM:ResetInstances();
	end
	
	local majorCivsAtPeace:table, majorCivsEmbassy:table, majorCivsAllied:table = GetDiplomaticVictoryMajorCivsInfo(currentPlayer);

	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		local playerID:number = player:GetID();
		
		if (playerID ~= currentPlayer:GetID() and currentPlayer:GetDiplomacy():HasMet(playerID)) then

			local majorCivInstance = majorCivsIM:GetInstance();
			ColorCivIcon(majorCivInstance, playerID);
			
			local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);
			textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, SIZE_RELIGION_CIV_ICON);
			
			--majorCivInstance.MajorCivPeace:SetHide(true);
			--majorCivInstance.MajorCivEmbassy:SetHide(true);
			--majorCivInstance.MajorCivAlliance:SetHide(true);
			
			local tooltip:string = "[NEWLINE]";
			
			if (IsPlayerInCivsTable(playerID, majorCivsAtPeace)) then
				--majorCivInstance.MajorCivPeace:SetHide(false);
				tooltip = tooltip .. GetRequirementIconStateText(true) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. ".[NEWLINE]";
			else
				tooltip = tooltip .. GetRequirementIconStateText(false) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. ".[NEWLINE]";
			end
			
			if (IsPlayerInCivsTable(playerID, majorCivsEmbassy)) then
				--majorCivInstance.MajorCivEmbassy:SetHide(false);
				tooltip = tooltip .. GetRequirementIconStateText(true) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_EMBASSIES_TEXT") .. ".[NEWLINE]";
			else
				tooltip = tooltip .. GetRequirementIconStateText(false) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_EMBASSIES_TEXT") .. ".[NEWLINE]";
			end
			
			if (IsPlayerInCivsTable(playerID, majorCivsAllied)) then
				--majorCivInstance.MajorCivAlliance:SetHide(false);
				tooltip = tooltip .. GetRequirementIconStateText(true) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_ALLIANCES_TEXT") .. ".";
			else
				tooltip = tooltip .. GetRequirementIconStateText(false) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_ALLIANCES_TEXT") .. ".";
			end
			
			if (textureSheet ~= nil) then
				majorCivInstance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
				SetLeaderTooltip(majorCivInstance, playerID, tooltip);
			end
		end
	end
	
	PopulateDiplomaticMajorCivsLabel(instance, majorCivsAtPeace, majorCivsEmbassy, majorCivsAllied);
end

function PopulateDiplomaticWorldCouncilLabel(instance:table, player:table)

	local hasWorldCouncil:boolean = GetDiplomaticVictoryHasWorldCouncil(player);
	local text:string = GetRequirementIconStateText(hasWorldCouncil) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_REQUIREMENT_WORLD_COUNCIL");

	instance.WorldCouncilLabel:SetText(text);
end

function PopulateDiplomaticMajorCivsLabel(instance:table, majorCivsAtPeace:table, majorCivsEmbassy:table, majorCivsAllied:table)

	local numMajorCivs:number = GetNumberOfItemsInCivsTable(PlayerManager:GetAliveMajors());
	local numMajorAlliesNeeded:number = DIPLOMATIC_MINIMUM_ALLIES_REQUIRED;
	local numMajorCivsAtPeace:number = GetNumberOfItemsInCivsTable(majorCivsAtPeace);
	local numMajorCivsEmbassy:number = GetNumberOfItemsInCivsTable(majorCivsEmbassy);
	local numMajorCivsAllied:number = GetNumberOfItemsInCivsTable(majorCivsAllied);
	
	local hasMajorCivRequirements:boolean = false;
	
	numMajorCivs = numMajorCivs - 1; -- do not include current player in count
	
	if (numMajorCivsAtPeace >= numMajorCivs and numMajorCivsEmbassy >= numMajorCivs and numMajorCivsAllied >= numMajorAlliesNeeded) then
		hasMajorCivRequirements = true;
	end

	local text:string = GetRequirementIconStateText(hasMajorCivRequirements);

	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_MAJOR_CIVS_TEXT") .. ": ";
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. " (" .. numMajorCivsAtPeace .. "/" .. numMajorCivs .. "). ";
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_EMBASSIES_TEXT") .. " (" .. numMajorCivsEmbassy .. "/" .. numMajorCivs .. "). ";
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_ALLIANCES_TEXT") .. " (" .. numMajorCivsAllied .. "/" .. numMajorAlliesNeeded .. ").";
	
	instance.MajorCivsLabel:SetText(text);
end

function PopulateDiplomaticMinorCivsInstance(instance:table, currentPlayer:table)

	local minorCivsIM:table = instance["MinorCivsIM"];
	
	if (minorCivsIM == nil) then
		minorCivsIM = InstanceManager:new("MinorCivsInstance", "CivIconBacking", instance.MinorCivsStack);
		instance["MinorCivsIM"] = minorCivsIM;
	end
	
	if (minorCivsIM ~= nil) then
		minorCivsIM:ResetInstances();
	end
	
	local minorCivsAtPeace:table, minorCivsSuzerain:table = GetDiplomaticVictoryMinorCivsInfo(currentPlayer);

	for i, player in ipairs(PlayerManager:GetAliveMinors()) do
		local playerID:number = player:GetID();
		
		if (playerID ~= currentPlayer:GetID() and currentPlayer:GetDiplomacy():HasMet(playerID)) then

			local minorCivInstance = minorCivsIM:GetInstance();
			ColorCivIcon(minorCivInstance, playerID);
			
			local civName:string, civIcon:string = GetCivNameAndIcon(playerID, true);
			textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(civIcon, SIZE_RELIGION_CIV_ICON);
			
			--minorCivInstance.MinorCivPeace:SetHide(true);
			--minorCivInstance.MinorCivSuzerain:SetHide(true);
			
			local tooltip:string = "[NEWLINE][NEWLINE]";
			
			if (IsPlayerInCivsTable(playerID, minorCivsAtPeace)) then
				--minorCivInstance.MinorCivPeace:SetHide(false);
				tooltip = tooltip .. GetRequirementIconStateText(true) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. ".[NEWLINE]";
			else
				tooltip = tooltip .. GetRequirementIconStateText(false) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. ".[NEWLINE]";
			end
			
			if (IsPlayerInCivsTable(playerID, minorCivsSuzerain)) then
				--minorCivInstance.MinorCivSuzerain:SetHide(false);
				tooltip = tooltip .. GetRequirementIconStateText(true) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_SUZERAIN_TEXT") .. ".";
			else
				tooltip = tooltip .. GetRequirementIconStateText(false) .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_SUZERAIN_TEXT") .. ".";
			end
			
			if (textureSheet ~= nil) then
				minorCivInstance.CivIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);
				SetCityStateTooltip(minorCivInstance, playerID, tooltip);
			end
		end
	end
	
	PopulateDiplomaticMinorCivsLabel(instance, minorCivsAtPeace, minorCivsSuzerain);
end

function SetCityStateTooltip(instance:table, playerID:number, tooltip:string)

	local player:table = Players[playerID];
	local playerConfig:table = PlayerConfigurations[playerID];
	
	if (m_LocalPlayer:GetDiplomacy():HasMet(playerID)) then
		instance.CivIcon:SetToolTipString(Locale.Lookup(playerConfig:GetPlayerName()) .. tooltip);
	else
		instance.CivIcon:SetToolTipString(Locale.Lookup("LOC_DIPLOPANEL_UNMET_PLAYER") .. tooltip);
	end
	
end

function PopulateDiplomaticMinorCivsLabel(instance:table, minorCivsAtPeace:table, minorCivsSuzerain:table)

	local numMinorCivs:number = GetNumberOfItemsInCivsTable(PlayerManager:GetAliveMinors());
	local numSuzerainNeeded:number = DIPLOMATIC_MINIMUM_SUZERAIN_REQUIRED;
	local numMinorCivsAtPeace:number = GetNumberOfItemsInCivsTable(minorCivsAtPeace);
	local numMinorCivsSuzerain:number = GetNumberOfItemsInCivsTable(minorCivsSuzerain);
	
	local hasMinorCivsRequirements:boolean = false;
	
	if (numMinorCivsAtPeace >= numMinorCivs and numMinorCivsSuzerain >= numSuzerainNeeded) then
		hasMinorCivsRequirements = true;
	end
	
	local text:string = GetRequirementIconStateText(hasMinorCivsRequirements);
	
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_MINOR_CIVS_TEXT") .. ": ";
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_PEACE_TEXT") .. " (" .. numMinorCivsAtPeace .. "/" .. numMinorCivs .. "). ";
	text = text .. Locale.Lookup("LOC_WORLD_RANKINGS_DIPLOMATIC_SUZERAIN_TEXT") .. " (" .. numMinorCivsSuzerain .. "/" .. numSuzerainNeeded .. ").";

	instance.MinorCivsLabel:SetText(text);
end

function RealizeDiplomaticStackSize()
	local _, screenY:number = UIManager:GetScreenSizeVal();

	if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
		local headerHeight:number = m_ActiveHeader[DATA_FIELD_HEADER_HEIGHT];
		
		m_ActiveHeader.AdvisorTextCentered:SetOffsetY(OFFSET_SCIENCE_REQUIREMENTS_Y + m_ActiveHeader.AdvisorText:GetSizeY() + PADDING_HEADER);
		headerHeight = headerHeight + m_ActiveHeader.AdvisorTextCentered:GetSizeY() + (PADDING_HEADER * 2);
		m_ActiveHeader.AdvisorTextBG:SetSizeY(headerHeight);
		m_ActiveHeader.AdvisorIcon:SetOffsetY(OFFSET_ADVISOR_ICON_Y + headerHeight);
		m_ActiveHeader.HeaderFrame:SetSizeY(OFFSET_ADVISOR_TEXT_Y + headerHeight);
		m_ActiveHeader.ContractHeaderButton:SetOffsetY(OFFSET_CONTRACT_BUTTON_Y + headerHeight);
		
		Controls.DiplomaticViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS + headerHeight + PADDING_HEADER);
		Controls.DiplomaticViewScrollbar:SetSizeY(screenY - (SIZE_STACK_DEFAULT + (headerHeight + PADDING_HEADER)));
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(false);
	else
		Controls.DiplomaticViewContents:SetOffsetY(OFFSET_VIEW_CONTENTS);
		Controls.DiplomaticViewScrollbar:SetSizeY(screenY - SIZE_STACK_DEFAULT);
		
		m_ActiveHeader.AdvisorTextCentered:SetHide(true);
	end

	RealizeStackAndScrollbar(Controls.DiplomaticViewStack, Controls.DiplomaticViewScrollbar, true);
end

-- // Diplomatic victory support functions
-- // -------------------------------------------------------------------------------------------
function GetDiplomaticVictoryProgressRatio(player:table)
	
	local majorRequirementsMet:boolean, majorCivsScore:number = GetDiplomaticVictoryMajorRequirementsCompleted(player);
	local minorRequirementsMet:boolean, minorCivsScore:number = GetDiplomaticVictoryMinorRequirementsCompleted(player);
	
	return Round((majorCivsScore + minorCivsScore) / 2, 1);
end

function GetDiplomaticVictoryDiplomaticPower(player:table)
	
	local majorRequirementsMet:boolean, majorCivsScore:number = GetDiplomaticVictoryMajorRequirementsCompleted(player);
	local minorRequirementsMet:boolean, minorCivsScore:number = GetDiplomaticVictoryMinorRequirementsCompleted(player);
	
	return (majorCivsScore + minorCivsScore) * 100;
end

function GetDiplomaticVictoryHasWorldCouncil(player:table)

	local projectCompleted:boolean = false;

	if (player ~= nil) then
		projectCompleted = HasPlayerCompletedProject(player, "PROJECT_WORLD_COUNCIL");
	end

	return projectCompleted;
end

function IsPlayerInCivsTable(currentPlayerID:number, civs:table)

	local result:boolean = false;

	if (civs ~= nil) then
		for c, player in ipairs(civs) do
			if (player:GetID() == currentPlayerID) then
				result = true; -- player exists in table
				break;
			end
		end
	end
	
	return result;
end

function GetNumberOfItemsInCivsTable(civs:table)

	local counter:number = 0;

	for i, player in ipairs(civs) do
		counter = counter + 1;
	end
	
	return counter;
end

function GetDiplomaticVictoryMajorCivsInfo(currentPlayer:table)

	local currentPlayerID:number = currentPlayer:GetID();
	local currentPlayerDiplomacy:table = currentPlayer:GetDiplomacy();
	local currentPlayerDiplomaticAI:table = currentPlayer:GetDiplomaticAI();

	local majorCivsAtPeace:table = {};
	local majorCivsEmbassy:table = {};
	local majorCivsAllied:table = {};
	
	if (currentPlayerDiplomacy ~= nil and currentPlayerDiplomaticAI ~= nil) then
	
		for i, player in ipairs(PlayerManager:GetAliveMajors()) do
			local playerID:number = player:GetID();
			
			if (playerID ~= currentPlayerID and currentPlayerDiplomacy:HasMet(playerID)) then

				-- if the players are at peace then we add a record to the majorCivsAtPeace table
				if (not currentPlayerDiplomacy:IsAtWarWith(playerID)) then
					table.insert(majorCivsAtPeace, player);
				end

				-- if the players have an embassy then add a record to the majorCivsEmbassy table
				if (currentPlayerDiplomacy:HasEmbassyAt(playerID)) then
					table.insert(majorCivsEmbassy, player);
				end
				
				-- if the players are allies then add a record to the majorCivsAllied table
				local relationshipStateID:number = currentPlayerDiplomaticAI:GetDiplomaticStateIndex(playerID);
				
				if (relationshipStateID ~= -1) then
					if (GameInfo.DiplomaticStates[relationshipStateID].StateType == "DIPLO_STATE_ALLIED") then
						table.insert(majorCivsAllied, player);
					end
				end
			end
		end
		
	end

	return majorCivsAtPeace, majorCivsEmbassy, majorCivsAllied;
end

function GetDiplomaticVictoryMinorCivsInfo(currentPlayer:table)

	local currentPlayerID:number = currentPlayer:GetID();
	local currentPlayerDiplomacy:table = currentPlayer:GetDiplomacy();

	local minorCivsAtPeace:table = {};
	local minorCivsSuzerain:table = {};
	
	if (currentPlayerDiplomacy ~= nil) then
	
		for i, player in ipairs(PlayerManager:GetAliveMinors()) do
			local playerID:number = player:GetID();
			
			if (playerID ~= currentPlayerID and currentPlayerDiplomacy:HasMet(playerID)) then
			
				-- if the players are at peace then we add a record to the minorCivsAtPeace table
				if (not currentPlayerDiplomacy:IsAtWarWith(playerID)) then
					table.insert(minorCivsAtPeace, player);
				end
			
				-- if the current player is the suzerain then add a record minorCivsSuzerain table
				local playerInfluence:table = player:GetInfluence();
				local suzerainID:number = -1;
				
				if (playerInfluence ~= nil) then
					suzerainID = playerInfluence:GetSuzerain();
				end
				
				if (suzerainID == currentPlayerID) then
					table.insert(minorCivsSuzerain, player);
				end
			end
		end
	
	end

	return minorCivsAtPeace, minorCivsSuzerain;
end

function GetDiplomaticVictoryMajorRequirementsCompleted(player:table)

	local majorCivsAtPeace:table, majorCivsEmbassy:table, majorCivsAllied:table = GetDiplomaticVictoryMajorCivsInfo(player);

	local numMajorCivs:number = GetNumberOfItemsInCivsTable(PlayerManager:GetAliveMajors());
	local numMajorAlliesNeeded:number = DIPLOMATIC_MINIMUM_ALLIES_REQUIRED;
	local numMajorCivsAtPeace:number = GetNumberOfItemsInCivsTable(majorCivsAtPeace);
	local numMajorCivsEmbassy:number = GetNumberOfItemsInCivsTable(majorCivsEmbassy);
	local numMajorCivsAllied:number = GetNumberOfItemsInCivsTable(majorCivsAllied);
	
	local hasMajorCivRequirements:boolean = false;
	
	numMajorCivs = numMajorCivs - 1; -- we do not include the current player
	
	if (numMajorCivsAtPeace >= numMajorCivs and numMajorCivsEmbassy >= numMajorCivs and numMajorCivsAllied >= numMajorAlliesNeeded) then
		hasMajorCivRequirements = true;
	end
	
	local maximumMajorCivsScore:number = (numMajorCivs * 2) + numMajorAlliesNeeded;
	local majorCivsScore:number = numMajorCivsAtPeace + numMajorCivsEmbassy + numMajorCivsAllied;
	local majorCivsProgressRatio:number = majorCivsScore / maximumMajorCivsScore;
	
	return hasMajorCivRequirements, majorCivsProgressRatio;
end

function GetDiplomaticVictoryMinorRequirementsCompleted(player:table)

	local minorCivsAtPeace:table, minorCivsSuzerain:table = GetDiplomaticVictoryMinorCivsInfo(player);

	local numMinorCivs:number = GetNumberOfItemsInCivsTable(PlayerManager:GetAliveMinors());
	local numSuzerainNeeded:number = DIPLOMATIC_MINIMUM_SUZERAIN_REQUIRED;
	local numMinorCivsAtPeace:number = GetNumberOfItemsInCivsTable(minorCivsAtPeace);
	local numMinorCivsSuzerain:number = GetNumberOfItemsInCivsTable(minorCivsSuzerain);
	
	local hasMinorCivsRequirements:boolean = false;
	
	if (numMinorCivsAtPeace >= numMinorCivs and numMinorCivsSuzerain >= numSuzerainNeeded) then
		hasMinorCivsRequirements = true;
	end
	
	local maximumMinorCivsScore:number = numMinorCivs + numSuzerainNeeded;
	local minorCivsScore:number = numMinorCivsAtPeace + numMinorCivsSuzerain;
	local minorCivsProgressRatio:number = minorCivsScore / maximumMinorCivsScore;

	return hasMinorCivsRequirements, minorCivsProgressRatio;
end

function GetDiplomaticVictoryCompleted(player:table)

	local result:boolean = false;

	-- has the player completed the world council project?
	if (GetDiplomaticVictoryHasWorldCouncil(player)) then
		local majorRequirementsMet:boolean, majorRequirementsRatio:number = GetDiplomaticVictoryMajorRequirementsCompleted(player);
		
		-- has the player met all the major civs requirements?
		if (majorRequirementsMet) then
			local minorRequirementsMet:boolean, minorRequirementsRatio:number = GetDiplomaticVictoryMinorRequirementsCompleted(player);
		
			-- has the player met all the minor civs requirements?
			result = minorRequirementsMet;
		end
	end
	
	return result;
end

-- // -------------------------------------------------------------------------------------------

-- ===========================================================================
--	Logic that governs extra tabs
-- ===========================================================================
function ToggleExtraTabs()
	local shouldHide:boolean = not Controls.ExtraTabs:IsHidden();
	Controls.ExtraTabs:SetHide(shouldHide);
	Controls.ExpandExtraTabs:SetSelected(not shouldHide);
end

function CloseExtraTabs()
	Controls.ExtraTabs:SetHide(true);
	Controls.ExpandExtraTabs:SetSelected(false);
	for _,tabInst in pairs(m_ExtraTabs) do
		tabInst.Button:SetSelected(false);
	end
end

-- ===========================================================================
--	Update player data and refresh the display state
-- ===========================================================================
function UpdatePlayerData()
	if (Game.GetLocalPlayer() ~= -1) then
		m_LocalPlayer = Players[Game.GetLocalPlayer()];
		m_LocalPlayerID = m_LocalPlayer:GetID();
	end
end
function UpdateData()
	UpdatePlayerData();
	if(m_LocalPlayer ~= nil and m_ActiveViewUpdate ~= nil) then
		m_ActiveViewUpdate();
	end
end

-- ===========================================================================
--	SCREEN EVENTS
-- ===========================================================================
function Open()
	UpdateData();
	m_AnimSupport.Show();
	UI.PlaySound("CityStates_Panel_Open");
	-- Ensure we're the only partial screen currenly up
	LuaEvents.WorldRankings_CloseCityStates();
end

-- ===========================================================================
function Close()	
	m_AnimSupport.Hide();
    if not ContextPtr:IsHidden() then
        UI.PlaySound("CityStates_Panel_Close");
    end
	LuaEvents.WorldRankings_Close();

	-- Don't reset lens if activating a modal lens. MinimapPanel will handle activating the proper lens.
	if UI.GetInterfaceMode() ~= InterfaceModeTypes.VIEW_MODAL_LENS then
		ResetTourismLens();
	end
end

-- ===========================================================================
function Toggle()	
	if(m_AnimSupport.IsVisible()) then
		Close();
	else
		Open();
	end
end

-- ===========================================================================
--	HOT-RELOADING EVENTS
-- ===========================================================================
function OnInit(isReload:boolean)
	if isReload then
		LuaEvents.GameDebug_GetValues(RELOAD_CACHE_ID);
	end
end

function OnShutdown()
	LuaEvents.GameDebug_AddValue(RELOAD_CACHE_ID, "isHidden", ContextPtr:IsHidden());
	if(m_TabSupport ~= nil and m_TabSupport.selectedControl ~= nil) then
		LuaEvents.GameDebug_AddValue(RELOAD_CACHE_ID, "selectedTabText", m_TabSupport.selectedControl:GetTextControl():GetText());
	end
	if(m_ActiveHeader ~= nil) then
		LuaEvents.GameDebug_AddValue(RELOAD_CACHE_ID, "activeHeaderExpanded", m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]);
	end
end

function OnGameDebugReturn(context:string, contextTable:table)
	if context == RELOAD_CACHE_ID and contextTable["isHidden"] ~= nil and not contextTable["isHidden"] then
		Open();
		-- Select previously selected tab
		local selectedTabText:string = contextTable["selectedTabText"];
		for _, tab in pairs(m_TabSupport.tabControls) do
			if tab:GetTextControl():GetText() == selectedTabText then
				m_TabSupport.SelectTab(tab);
			end
		end

		if(m_ActiveHeader ~= nil) then
			m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED] = contextTable["activeHeaderExpanded"];
			if(m_ActiveHeader[DATA_FIELD_HEADER_EXPANDED]) then
				OnExpandHeader(0, 0, m_ActiveHeader.ExpandHeaderButton);
			else
				OnContractHeader(0, 0, m_ActiveHeader.ContractHeaderButton);
			end
		end
	end
end

-- ===========================================================================
--	Hot-seat functionality
-- ===========================================================================
function OnLocalPlayerTurnEnd()
	if(GameConfiguration.IsHotseat()) then
		Close();
	end
end

-- // ------------------------------------------------------------------------
-- // CiVI Reformation victories game event handler
-- // ------------------------------------------------------------------------
function OnLocalPlayerTurnBegin()

	local victoryAchieved:boolean = false;
	local victoryType:string = "";
	
	-- check if the player has achieved any of the custom victory conditions
	victoryType = GetCustomVictoryCompleted(m_LocalPlayer);
	
	if (victoryType ~= "") then
		LuaEvents.CustomVictoryTriggered(m_LocalPlayer, victoryType);
	end
	
	-- check if any other player or ai have achieved any custom victory type
	for i, player in ipairs(PlayerManager:GetAliveMajors()) do
		if (player:GetID() ~= m_LocalPlayer:GetID()) then
			victoryType = GetCustomVictoryCompleted(player);
			
			if (victoryType ~= "") then
				LuaEvents.CustomVictoryTriggered(m_LocalPlayer, "DEFEAT_DEFAULT");
				break;
			end
		end
	end

end

function GetCustomVictoryCompleted(player:table)

	local victoryType:string = "";

	if (GetEconomicVictoryCompleted(player)) then
		victoryType = "VICTORY_ECONOMIC";
		
	elseif (GetMetropolisVictoryCompleted(player, METROPOLIS_MIN_HAPPINESS)) then
		victoryType = "VICTORY_METROPOLIS";
		
	elseif (GetTerritorialVictoryCompleted(player)) then
		victoryType = "VICTORY_TERRITORIAL";
	
	elseif (GetDiplomaticVictoryCompleted(player)) then
		victoryType = "VICTORY_DIPLOMATIC";
	
	end
	
	return victoryType;
end

-- // ------------------------------------------------------------------------

-- ===========================================================================
--	Game Events
-- ===========================================================================
function OnCapitalCityChanged()
	if(m_AnimSupport.IsVisible()) then
		m_AnimSupport.Hide();
	end
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnInterfaceModeChanged(eOldMode:number, eNewMode:number)
	if m_AnimSupport.IsVisible() and eNewMode == InterfaceModeTypes.VIEW_MODAL_LENS then
		Close();
	end
end

-- ===========================================================================
--	LUA Event
--	Explicit close (from partial screen hooks), part of closing everything,
-- ===========================================================================
function OnCloseAllExcept( contextToStayOpen:string )
	if contextToStayOpen == ContextPtr:GetID() then return; end
	Close();
end

-- ===========================================================================
--	INIT (Generic)
-- ===========================================================================
function Initialize()

	-- Animation Controller
	m_AnimSupport = CreateScreenAnimation(Controls.SlideAnim, Close);

	-- Hot-Reload Events
	ContextPtr:SetInitHandler(OnInit);
	ContextPtr:SetShutdown(OnShutdown);
	LuaEvents.GameDebug_Return.Add(OnGameDebugReturn);

	-- UI Callbacks
	ContextPtr:SetInputHandler(m_AnimSupport.OnInputHandler, true);

	Controls.CloseButton:SetOffsetY(50);		-- Move close button down to account for TabHeader
	Controls.CloseButton:RegisterCallback(Mouse.eLClick, Close);
	Controls.CloseButton:RegisterCallback( Mouse.eMouseEnter, function() UI.PlaySound("Main_Menu_Mouse_Over"); end);
	Controls.ExpandExtraTabs:RegisterCallback(Mouse.eLClick, ToggleExtraTabs);
	Controls.ScoreDetailsCheck:RegisterCallback(Mouse.eLClick, ToggleScoreDetails);
	Controls.ScoreDetailsButton:RegisterCallback(Mouse.eLClick, ToggleScoreDetails);
	Controls.ScoreDetailsButton:RegisterCallback( Mouse.eMouseEnter, function() UI.PlaySound("Main_Menu_Mouse_Over"); end);
	Controls.Title:SetText(Locale.Lookup("LOC_WORLD_RANKINGS_TITLE"));
	Controls.TabHeader:ChangeParent(Controls.Background);		-- To make it render beneath the banner image

	-- Game Events
	Events.CapitalCityChanged.Add(OnCapitalCityChanged);
	Events.InterfaceModeChanged.Add( OnInterfaceModeChanged );
	Events.LocalPlayerTurnBegin.Add(OnLocalPlayerTurnBegin); -- // added for CiVI Reformation victories
	Events.LocalPlayerTurnEnd.Add(OnLocalPlayerTurnEnd);
	Events.SystemUpdateUI.Add(m_AnimSupport.OnUpdateUI);

	-- Lua Events
	LuaEvents.PartialScreenHooks_ToggleWorldRankings.Add(Toggle);
	LuaEvents.PartialScreenHooks_OpenWorldRankings.Add(Open);
	LuaEvents.PartialScreenHooks_CloseWorldRankings.Add(Close);
	LuaEvents.PartialScreenHooks_CloseAllExcept.Add(OnCloseAllExcept);
	
	-- // CiVI Reformation Victories: load the collection of luxury resources that have spawned on the map
	LoadMapLuxuryResourceData();
	
	if Game.GetLocalPlayer() ~= -1 then
		UpdatePlayerData();
		PopulateTabs();
	end
end
Initialize();
